from typing import Sequence
import numpy as np
import numba
import regridding
from regridding import _util


def _weights_multilinear(
    coordinates_input: tuple[np.ndarray, ...],
    coordinates_output: tuple[np.ndarray, ...],
    axis_input: None | int | Sequence[int] = None,
    axis_output: None | int | Sequence[int] = None,
) -> tuple[np.ndarray, tuple[int, ...], tuple[int, ...]]:
    indices_output = regridding.find_indices(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        axis_input=axis_input,
        axis_output=axis_output,
    )
    result = _weights_from_indices_multilinear(
        indices_output=indices_output,
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        axis_input=axis_input,
        axis_output=axis_output,
    )
    return result


def _weights_from_indices_multilinear(
    indices_output: tuple[np.ndarray, ...],
    coordinates_input: tuple[np.ndarray, ...],
    coordinates_output: tuple[np.ndarray, ...],
    axis_input: None | int | Sequence[int] = None,
    axis_output: None | int | Sequence[int] = None,
) -> tuple[np.ndarray, tuple[int, ...], tuple[int, ...]]:
    (
        coordinates_input,
        coordinates_output,
        axis_input,
        axis_output,
        shape_input,
        shape_output,
        shape_orthogonal,
    ) = _util._normalize_input_output_coordinates(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        axis_input=axis_input,
        axis_output=axis_output,
    )

    axis_input_numba = ~np.arange(len(axis_input))[::-1]
    axis_output_numba = ~np.arange(len(axis_output))[::-1]

    shape_input_numba = tuple(shape_input[ax] for ax in axis_input)
    shape_output_numba = tuple(shape_output[ax] for ax in axis_output)

    indices_output = tuple(
        np.moveaxis(v, axis_output, axis_output_numba).reshape(-1, *shape_output_numba)
        for v in indices_output
    )
    coordinates_input = tuple(
        np.moveaxis(v, axis_input, axis_input_numba).reshape(-1, *shape_input_numba)
        for v in coordinates_input
    )
    coordinates_output = tuple(
        np.moveaxis(v, axis_output, axis_output_numba).reshape(-1, *shape_output_numba)
        for v in coordinates_output
    )

    if len(axis_input) == 1:
        weights_list = _weights_from_indices_multilinear_1d(
            indices_output=indices_output,
            coordinates_input=coordinates_input,
            coordinates_output=coordinates_output,
        )
    else:
        raise ValueError(
            f"{len(axis_input)}-dimensional multilinear interpolation is not supported"
        )

    num_d = len(weights_list)
    weights = np.empty(shape=num_d, dtype=numba.typed.List)
    for d in range(num_d):
        weights[d] = weights_list[d]
    weights = weights.reshape(shape_orthogonal)

    return weights, shape_input, shape_output


@numba.njit(parallel=False)
def _weights_from_indices_multilinear_1d(
    indices_output: tuple[np.ndarray],
    coordinates_input: tuple[np.ndarray],
    coordinates_output: tuple[np.ndarray],
) -> np.ndarray:
    (i_output,) = indices_output
    (x_input,) = coordinates_input
    (x_output,) = coordinates_output

    num_d, num_i_input = x_input.shape
    num_d, num_i_output = x_output.shape

    weights = numba.typed.List()

    for d in range(num_d):
        weights_d = numba.typed.List()
        for _ in range(0):
            weights_d.append((0.0, 0.0, 0.0))

        for i in numba.prange(num_i_output):
            i0 = i_output[d, i]
            i1 = i0 + 1

            x0 = x_input[d, i0]
            x1 = x_input[d, i1]
            x = x_output[d, i]

            w1 = (x - x0) / (x1 - x0)
            w0 = 1 - w1

            weights_d.append((i0, i, w0))
            weights_d.append((i1, i, w1))

        weights.append(weights_d)

    return weights



import numpy as np
import regridding

coordinates_input = None
coordinates_output = None


def setup_find_indices_1d(num: int, method: str):
    global coordinates_input
    global coordinates_output
    coordinates_input = (np.linspace(-1, 1, num=num),)
    coordinates_output = (np.linspace(-1.1, 1.1, num=num),)
    regridding.find_indices(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        method=method,
    )


def time_find_indices_1d(num: int, method: str):
    regridding.find_indices(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        method=method,
    )


time_find_indices_1d.setup = setup_find_indices_1d
time_find_indices_1d.params = (
    list(np.linspace(0, 1e4, num=11, dtype=int)[1:]),
    ["brute", "searchsorted"],
)
time_find_indices_1d.param_names = ["num", "method"]



import numpy as np
import regridding

coordinates_input = None
coordinates_output = None
values_input = None


def setup_regrid_2d(
    num: int,
    method: str,
):
    num_x = num_y = num

    x = np.linspace(-1, 1, num=num_x)
    y = np.linspace(-1, 1, num=num_y)
    x, y = np.meshgrid(x, y, indexing="ij")

    angle = 0.4
    x_input = x * np.cos(angle) - y * np.sin(angle) + 0.05 * x * x
    y_input = x * np.sin(angle) + y * np.cos(angle) + 0.05 * y * y

    x_output = np.linspace(x_input.min(), x_input.max(), num=num_x)
    y_output = np.linspace(y_input.min(), y_input.max(), num=num_y)
    x_output, y_output = np.meshgrid(x_output, y_output, indexing="ij")

    global coordinates_input
    global coordinates_output
    global values_input

    coordinates_input = (x_input, y_input)
    coordinates_output = (x_output, y_output)
    values_input = np.random.uniform(0, 1, size=(num_x - 1, num_y - 1))

    regridding.regrid(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        values_input=values_input,
        method=method,
    )


def time_regrid_2d(
    num: int,
    method: str,
):
    regridding.regrid(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        values_input=values_input,
        method=method,
    )


time_regrid_2d.setup = setup_regrid_2d
time_regrid_2d.params = (
    [100, 200, 300, 400, 500],
    ["conservative"],
)
time_regrid_2d.param_names = (
    "number of edges per axis",
    "method",
)
time_regrid_2d.timeout = 480



from __future__ import annotations
import math
import numpy as np
import numba

__all__ = [
    "ndarray_linear_interpolation",
]


def ndarray_linear_interpolation(
    a: np.ndarray,
    indices: tuple[np.ndarray, ...],
    axis: None | int | tuple[int, ...] = None,
    axis_indices: None | int | tuple[int, ...] = None,
):
    """
    Interpolate a :class:`numpy.ndarray` onto a new grid.

    Similar to :func:`scipy.ndimage.map_coordinates`, but allows for interpolation along only some of the
    axes of ``a``.

    Parameters
    ----------
    a
        The input array to be interpolated. Should have at least as many axes as ``len(indices)``
    indices
        The new indices where ``a`` will be evaluated. The number of indices, ``len(indices)`` should be equal
        to the number of interpolation axes, ``len(axis)``
    axis
        The axes of ``a`` to interpolate. If :class:`None`, interpolate over all the axes of ``a``.
    axis_indices
        The axes of ``indices`` that are complementary to the axes in ``axis``.
        If :class:`None`, all axes of indices are complementary to the axes in ``axis``.

    Returns
    -------

        The interpolated array

    Examples
    --------

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import regridding

        x = np.arange(5)
        y = np.arange(6)

    Interpolate a 1D array

    .. jupyter-execute::

        a = x * x

        x_new = np.linspace(0, x.size - 1, num=20)
        a_new = regridding.ndarray_linear_interpolation(a, indices=(x_new, ))

        plt.figure(figsize=(6, 3));
        plt.scatter(x, a, s=100, label="original", zorder=1);
        plt.scatter(x_new, a_new, label="interpolated", zorder=0);
        plt.legend();

    """

    shape_a = a.shape
    ndim_a = a.ndim

    indices_broadcasted = np.broadcast(*indices)
    shape_indices = indices_broadcasted.shape
    ndim_indices = indices_broadcasted.ndim

    if axis is None:
        axis = tuple(range(ndim_a))
    axis = np.core.numeric.normalize_axis_tuple(axis, ndim=ndim_a)
    axis = tuple(~(~np.array(axis) % ndim_a))

    if axis_indices is None:
        axis_indices = tuple(range(ndim_indices))
    axis_indices = np.core.numeric.normalize_axis_tuple(axis_indices, ndim=ndim_indices)
    axis_indices = tuple(~(~np.array(axis_indices) % ndim_indices))

    if len(indices) != len(axis):
        raise ValueError(
            f"The number of indices, {len(indices)}, must match the number of elements in axis, {len(axis)}"
        )

    axis_orthogonal_a = tuple(ax for ax in range(-ndim_a, 0) if ax not in axis)
    axis_orthogonal_indices = tuple(
        ax for ax in range(-ndim_indices, 0) if ax not in axis_indices
    )

    shape_orthogonal_a = tuple(shape_a[ax] for ax in axis_orthogonal_a)
    shape_orthogonal_indices = tuple(
        shape_indices[ax] for ax in axis_orthogonal_indices
    )

    shape_orthogonal = np.broadcast_shapes(shape_orthogonal_a, shape_orthogonal_indices)

    ndim_broadcasted_a = len(shape_orthogonal) + len(axis)
    ndim_broadcasted_indices = len(shape_orthogonal) + len(axis_indices)

    shape_broadcasted_a = tuple(
        shape_a[ax] if ax in axis else shape_orthogonal[axis_orthogonal_a.index(ax)]
        for ax in range(-ndim_broadcasted_a, 0)
    )
    shape_broadcasted_indices = tuple(
        (
            shape_indices[ax]
            if ax in axis_indices
            else shape_orthogonal[axis_orthogonal_indices.index(ax)]
        )
        for ax in range(-ndim_broadcasted_indices, 0)
    )

    a = np.broadcast_to(a, shape_broadcasted_a)
    indices = tuple(np.broadcast_to(ind, shape_broadcasted_indices) for ind in indices)

    result = np.empty(shape_broadcasted_indices)

    def index_a_indices(index: tuple[int, ...]) -> tuple[tuple, tuple]:
        index_a = tuple(
            slice(None) if ax in axis else index[axis_orthogonal_a.index(ax)]
            for ax in range(-ndim_broadcasted_a, 0)
        )
        index_indices = tuple(
            (
                slice(None)
                if ax in axis_indices
                else index[axis_orthogonal_indices.index(ax)]
            )
            for ax in range(-ndim_broadcasted_indices, 0)
        )
        return index_a, index_indices

    if len(axis) == 1:
        (x,) = indices

        for index in np.ndindex(*shape_orthogonal):
            index_a, index_indices = index_a_indices(index)
            result[index_indices] = _ndarray_linear_interpolation_1d(
                a=a[index_a],
                x=x[index_indices].reshape(-1),
            ).reshape(x[index_indices].shape)

    elif len(axis) == 2:
        x, y = indices

        for index in np.ndindex(*shape_orthogonal):
            index_a, index_indices = index_a_indices(index)
            result[index_indices] = _ndarray_linear_interpolation_2d(
                a=a[index_a],
                x=x[index_indices].reshape(-1),
                y=y[index_indices].reshape(-1),
            ).reshape(x[index_indices].shape)

    else:
        raise NotImplementedError

    return result


@numba.njit(cache=True, parallel=True)
def _ndarray_linear_interpolation_1d(
    a: np.ndarray,
    x: np.ndarray,
) -> np.ndarray:
    shape_output = x.shape
    (size_output,) = shape_output

    result = np.empty(size_output)

    for i in numba.prange(size_output):
        result[i] = _linear_interpolation(a, x[i])

    return result


@numba.njit(cache=True, parallel=True)
def _ndarray_linear_interpolation_2d(
    a: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
) -> np.ndarray:
    shape_output = x.shape
    (size_output,) = shape_output

    result = np.empty(size_output)

    for i in numba.prange(size_output):
        result[i] = _bilinear_interpolation(a, x[i], y[i])

    return result


@numba.njit(cache=True)
def _linear_interpolation(
    a: np.ndarray,
    x: float,
) -> float:
    (shape_input_x,) = a.shape

    x_0 = int(math.floor(x))

    if x_0 < 0:
        x_0 = 0
    elif x_0 > shape_input_x - 2:
        x_0 = shape_input_x - 2

    x_1 = x_0 + 1

    a_0 = a[x_0]
    a_1 = a[x_1]

    dx = x - x_0

    return a_0 * (1 - dx) + a_1 * dx


@numba.njit(cache=True)
def _bilinear_interpolation(
    a: np.ndarray,
    x: float,
    y: float,
) -> float:
    shape_input_x, shape_input_y = a.shape

    x_00 = int(math.floor(x))
    y_00 = int(math.floor(y))

    if x_00 < 0:
        x_00 = 0
    elif x_00 > shape_input_x - 2:
        x_00 = shape_input_x - 2

    if y_00 < 0:
        y_00 = 0
    elif y_00 > shape_input_y - 2:
        y_00 = shape_input_y - 2

    x_01 = x_00
    x_10 = x_11 = x_00 + 1

    y_01 = y_11 = y_00 + 1
    y_10 = y_00

    a_00 = a[x_00, y_00]
    a_01 = a[x_01, y_01]
    a_10 = a[x_10, y_10]
    a_11 = a[x_11, y_11]

    dx = x - x_00
    dy = y - y_00

    w_00 = (1 - dx) * (1 - dy)
    w_01 = (1 - dx) * dy
    w_10 = dx * (1 - dy)
    w_11 = dx * dy

    return (a_00 * w_00) + (a_01 * w_01) + (a_10 * w_10) + (a_11 * w_11)



from typing import Sequence
import numpy as np
import numba
from regridding import _util
from regridding._conservative_ramshaw import _conservative_ramshaw


def _weights_conservative(
    coordinates_input: tuple[np.ndarray, ...],
    coordinates_output: tuple[np.ndarray, ...],
    axis_input: None | int | Sequence[int] = None,
    axis_output: None | int | Sequence[int] = None,
) -> tuple[np.ndarray, tuple[int, ...], tuple[int, ...]]:
    (
        coordinates_input,
        coordinates_output,
        axis_input,
        axis_output,
        shape_input,
        shape_output,
        shape_orthogonal,
    ) = _util._normalize_input_output_coordinates(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        axis_input=axis_input,
        axis_output=axis_output,
    )

    shape_values_input = list(shape_input)
    for ax in axis_input:
        shape_values_input[ax] -= 1
    shape_values_input = tuple(shape_values_input)

    shape_values_output = list(shape_output)
    for ax in axis_output:
        shape_values_output[ax] -= 1
    shape_values_output = tuple(shape_values_output)

    weights = np.empty(shape_orthogonal, dtype=numba.typed.List)

    for index in np.ndindex(*shape_orthogonal):
        index_vertices_input = list(reversed(index))

        for ax in axis_input:
            index_vertices_input.insert(~ax, slice(None))
        index_vertices_input = tuple(reversed(index_vertices_input))

        index_vertices_output = list(reversed(index))
        for ax in axis_output:
            index_vertices_output.insert(~ax, slice(None))
        index_vertices_output = tuple(reversed(index_vertices_output))

        if len(axis_input) == 1:
            raise NotImplementedError("1D regridding not supported")

        elif len(axis_input) == 2:
            coordinates_input_x, coordinates_input_y = coordinates_input
            coordinates_output_x, coordinates_output_y = coordinates_output
            weights[index] = _conservative_ramshaw(
                grid_input=(
                    coordinates_input_x[index_vertices_input],
                    coordinates_input_y[index_vertices_input],
                ),
                grid_output=(
                    coordinates_output_x[index_vertices_output],
                    coordinates_output_y[index_vertices_output],
                ),
            )

        else:
            raise NotImplementedError(
                "Regridding operations greater than 2D are not supported"
            )

    return weights, shape_values_input, shape_values_output



import sys
import math
import numpy as np
import numba
# import matplotlib.pyplot as plt

__all__ = []


@numba.njit(
    error_model="numpy",
    # parallel=True,
    # inline="always",
    # boundscheck=True,
    # cache=True,
)
def _conservative_ramshaw(
        grid_input: tuple[np.ndarray, np.ndarray],
        grid_output: tuple[np.ndarray, np.ndarray],
        epsilon: float = 1e-10,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:

    weights = numba.typed.List()
    for x in range(0):
        weights.append((0., 0., 0.))

    input_x, input_y = grid_input

    shape_input = input_x.shape

    axes = 0, 1

    # k = slice(None, -1)
    # # k = slice(1, None)
    # grids_sweep = grids_sweep[k]
    # grids_static = grids_static[k]
    # grids_input = grids_input[k]
    # axes = 1,

    area_input = np.zeros((shape_input[0] - 1, shape_input[1] - 1))
    for axis in axes:
        _grid_area_sweep(
            out=area_input,
            grid_x=grid_input[0],
            grid_y=grid_input[1],
            axis=axis,
        )

    # values_input = values_input / area_input

    grid_static_x, grid_static_y = grid_output
    grid_sweep_x, grid_sweep_y = grid_input
    for axis in axes:
        _sweep_axis(
            area_input=area_input,
            grid_sweep_x=grid_sweep_x,
            grid_sweep_y=grid_sweep_y,
            grid_static_x=grid_static_x,
            grid_static_y=grid_static_y,
            axis=axis,
            grid_input="sweep",
            epsilon=epsilon,
            weights=weights,
        )

    grid_static_x, grid_static_y = grid_input
    grid_sweep_x, grid_sweep_y = grid_output
    for axis in axes:
        _sweep_axis(
            area_input=area_input,
            grid_sweep_x=grid_sweep_x,
            grid_sweep_y=grid_sweep_y,
            grid_static_x=grid_static_x,
            grid_static_y=grid_static_y,
            axis=axis,
            grid_input="static",
            epsilon=epsilon,
            weights=weights,
        )

    # return values_output
    return weights


@numba.njit(cache=True, error_model="numpy", parallel=True)
def _sweep_axis(
        # values_input: np.ndarray,
        # values_output: np.ndarray,
        area_input: np.ndarray,
        grid_static_x: np.ndarray,
        grid_static_y: np.ndarray,
        grid_sweep_x: np.ndarray,
        grid_sweep_y: np.ndarray,
        axis: int,
        grid_input: str,
        epsilon: float,
        weights: numba.typed.List,
) -> None:

    if grid_input == "static":
        input_is_sweep = False
    elif grid_input == "sweep":
        input_is_sweep = True
    else:
        print(f"The `grid_input` argument must have the value 'static' or 'sweep', got '{grid_input}'")

    # print(input_is_sweep)

    if axis == 0:
        grid_sweep_y, grid_sweep_x = grid_sweep_x.T, grid_sweep_y.T
        grid_static_y, grid_static_x = grid_static_x, grid_static_y
        # if input_is_sweep:
        #     values_input = values_input.T
        # else:
        #     values_output = values_output.T
    elif axis == 1:
        pass
    else:
        print(f"The `axis` argument must be 0 or 1, got {axis}")

    shape_static_x, shape_static_y = list(grid_static_x.shape)
    shape_sweep_x, shape_sweep_y = list(grid_sweep_x.shape)

    edges_bottom = slice(None, ~0), 0
    edges_right = ~0, slice(None, ~0)
    edges_top = slice(None, 0, -1), ~0
    edges_left = 0, slice(None, 0, -1)

    edges_border_static_x = np.concatenate((
        grid_static_x[edges_bottom],
        grid_static_x[edges_right],
        grid_static_x[edges_top],
        grid_static_x[edges_left],
    ))
    edges_border_static_y = np.concatenate((
        grid_static_y[edges_bottom],
        grid_static_y[edges_right],
        grid_static_y[edges_top],
        grid_static_y[edges_left],
    ))

    cells_bottom = slice(None), 0
    cells_right = ~0, slice(None)
    cells_top = slice(None, None, -1), ~0
    cells_left = 0, slice(None, None, -1)

    grid_static_m = np.arange(shape_static_x)[:, np.newaxis]
    grid_static_n = np.arange(shape_static_y)[np.newaxis, :]
    grid_static_m, grid_static_n = np.broadcast_arrays(grid_static_m, grid_static_n)
    edges_border_static_m = np.concatenate((
        grid_static_m[edges_bottom],
        grid_static_m[edges_right],
        grid_static_m[edges_top],
        grid_static_m[edges_left],
    ))
    edges_border_static_n = np.concatenate((
        grid_static_n[edges_bottom],
        grid_static_n[edges_right],
        grid_static_n[edges_top],
        grid_static_n[edges_left],
    ))

    cells_static_m = np.arange(shape_static_x - 1)[:, np.newaxis]
    cells_static_n = np.arange(shape_static_y - 1)[np.newaxis, :]
    cells_static_m, cells_static_n = np.broadcast_arrays(cells_static_m, cells_static_n)
    cells_border_static_m = np.concatenate((
        cells_static_m[cells_bottom],
        cells_static_m[cells_right],
        cells_static_m[cells_top],
        cells_static_m[cells_left],
    ))
    cells_border_static_n = np.concatenate((
        cells_static_n[cells_bottom],
        cells_static_n[cells_right],
        cells_static_n[cells_top],
        cells_static_n[cells_left],
    ))

    weight = numba.typed.List()

    for i in range(shape_sweep_x):
        w = numba.typed.List()
        for _ in range(0):
            w.append((0., 0., 0.))
        weight.append(w)

    for i in numba.prange(shape_sweep_x):
        i = numba.types.int64(i)

        sweep_is_inside_static = False
        j = 0

        point_sweep_1x = grid_sweep_x[i, j]
        point_sweep_1y = grid_sweep_y[i, j]
        point_sweep_2x = grid_sweep_x[i, j + 1]
        point_sweep_2y = grid_sweep_y[i, j + 1]

        if _point_is_inside_polygon(
                vertices_x=edges_border_static_x,
                vertices_y=edges_border_static_y,
                point_x=point_sweep_1x,
                point_y=point_sweep_1y,
                epsilon=epsilon,
        ):
            # print("initial point inside...")
            m, n = _indices_of_line_brute(
                grid_x=grid_static_x,
                grid_y=grid_static_y,
                point_1x=point_sweep_1x,
                point_1y=point_sweep_1y,
                point_2x=point_sweep_2x,
                point_2y=point_sweep_2y,
                epsilon=epsilon,
            )
            sweep_is_inside_static = True
        else:
            m = sys.maxsize
            n = sys.maxsize

        m_old = sys.maxsize
        n_old = sys.maxsize

        while True:

            point_sweep_2x = grid_sweep_x[i, j + 1]
            point_sweep_2y = grid_sweep_y[i, j + 1]

            # print("i", i)
            # print("j", j)
            # print("m", m)
            # print("n", n)
            # print("point_sweep_1", (point_sweep_1x, point_sweep_1y))
            # print("point_sweep_2", (point_sweep_2x, point_sweep_2y))
            #
            # plt.figure()
            # plt.plot(grid_static_x, grid_static_y, color="black")
            # plt.plot(grid_static_x.T, grid_static_y.T, color="black")
            # plt.plot(grid_sweep_x, grid_sweep_y, color="red")
            # plt.plot(grid_sweep_x.T, grid_sweep_y.T, color="red")
            # plt.plot([point_sweep_1x, point_sweep_2x], [point_sweep_1y, point_sweep_2y], zorder=10)

            if not sweep_is_inside_static:
                point_sweep_2x, point_sweep_2y, j_new, m_new, n_new, sweep_is_inside_static = _step_outside_static(
                    # area_output=area_output,
                    grid_static_x=grid_static_x,
                    grid_static_y=grid_static_y,
                    edges_border_static_m=edges_border_static_m,
                    edges_border_static_n=edges_border_static_n,
                    cells_border_static_m=cells_border_static_m,
                    cells_border_static_n=cells_border_static_n,
                    # shape_sweep_x=shape_sweep_x,
                    # shape_sweep_y=shape_sweep_y,
                    point_sweep_1x=point_sweep_1x,
                    point_sweep_1y=point_sweep_1y,
                    point_sweep_2x=point_sweep_2x,
                    point_sweep_2y=point_sweep_2y,
                    # input_is_sweep=input_is_sweep,
                    epsilon=epsilon,
                    # i=i,
                    j=j,
                    m=m,
                    n=n,
                )

            else:
                point_sweep_2x, point_sweep_2y, j_new, m_new, n_new = _step_inside_static(
                    # values_input=values_input,
                    # values_output=values_output,
                    # area_output=area_output,
                    area_input=area_input,
                    grid_static_x=grid_static_x,
                    grid_static_y=grid_static_y,
                    shape_sweep_x=shape_sweep_x,
                    shape_sweep_y=shape_sweep_y,
                    point_sweep_1x=point_sweep_1x,
                    point_sweep_1y=point_sweep_1y,
                    point_sweep_2x=point_sweep_2x,
                    point_sweep_2y=point_sweep_2y,
                    axis=axis,
                    input_is_sweep=input_is_sweep,
                    epsilon=epsilon,
                    m_old=m_old,
                    n_old=n_old,
                    i=i,
                    j=j,
                    m=m,
                    n=n,
                    weight=weight[i],
                )

            # print("j_new", j_new)
            # print("m_new", m_new)
            # print("n_new", n_new)

            # if input_is_sweep:
            #     plt.pcolormesh(grid_static_x, grid_static_y, values_output, zorder=-100)
            # else:
            #     plt.pcolormesh(grid_sweep_x, grid_sweep_y, values_output, zorder=-100)
            # plt.colorbar()

            if j_new >= (shape_sweep_y - 1):
                break
            if sweep_is_inside_static:
                if not 0 <= m_new < (shape_static_x - 1):
                    break
                if not 0 <= n_new < (shape_static_y - 1):
                    break

            j = j_new
            m_old = m
            n_old = n
            m = m_new
            n = n_new
            point_sweep_1x = point_sweep_2x
            point_sweep_1y = point_sweep_2y

            # print()

        # print("---------------------------")

    for i in range(shape_sweep_x):
        weight_i = weight[i]
        for w in range(len(weight_i)):
            weights.append(weight_i[w])

        # indices_input += index_input[i]
        # indices_output += index_output[i]
        # weights += weight[i]


@numba.njit(cache=True, inline="always")
def _step_outside_static(
        # area_output: np.ndarray,
        grid_static_x: np.ndarray,
        grid_static_y: np.ndarray,
        edges_border_static_m: np.ndarray,
        edges_border_static_n: np.ndarray,
        cells_border_static_m: np.ndarray,
        cells_border_static_n: np.ndarray,
        # shape_sweep_x: int,
        # shape_sweep_y: int,
        point_sweep_1x: float,
        point_sweep_1y: float,
        point_sweep_2x: float,
        point_sweep_2y: float,
        # input_is_sweep: bool,
        epsilon: float,
        # i: int,
        j: int,
        m: int,
        n: int,

) -> tuple[float, float, int, int, int, bool]:

    # print("step outside")

    direction_sweep_x = point_sweep_2x - point_sweep_1x
    direction_sweep_y = point_sweep_2y - point_sweep_1y

    sweep_is_inside_static = False
    j_new = j + 1
    m_new = m
    n_new = n

    u_min = np.inf

    # print("m and n not set")
    for v in range(len(edges_border_static_m)):

        index_vertex_1 = v - 1
        index_vertex_2 = v

        vertex_static_1m = edges_border_static_m[index_vertex_1]
        vertex_static_1n = edges_border_static_n[index_vertex_1]
        vertex_static_2m = edges_border_static_m[index_vertex_2]
        vertex_static_2n = edges_border_static_n[index_vertex_2]

        direction_m = vertex_static_2m - vertex_static_1m
        direction_n = vertex_static_2n - vertex_static_1n

        vertex_static_1x = grid_static_x[vertex_static_1m, vertex_static_1n]
        vertex_static_1y = grid_static_y[vertex_static_1m, vertex_static_1n]

        vertex_static_2x = grid_static_x[vertex_static_2m, vertex_static_2n]
        vertex_static_2y = grid_static_y[vertex_static_2m, vertex_static_2n]

        direction_static_x = vertex_static_2x - vertex_static_1x
        direction_static_y = vertex_static_2y - vertex_static_1y

        # plt.plot(
        #     [vertex_static_1x, vertex_static_2x],
        #     [vertex_static_1y, vertex_static_2y],
        #     color="cyan"
        # )

        t, u = _two_line_segment_intersection_parameters(
            x1=vertex_static_1x, y1=vertex_static_1y,
            x2=vertex_static_2x, y2=vertex_static_2y,
            x3=point_sweep_1x, y3=point_sweep_1y,
            x4=point_sweep_2x, y4=point_sweep_2y,
        )

        if u > u_min:
            continue

        if (0 - epsilon) < u < (1 + epsilon):
            if (0 - epsilon) < t < (1 + epsilon):

                sweep_is_inside_static = True

                m_new = cells_border_static_m[index_vertex_1]
                n_new = cells_border_static_n[index_vertex_1]

                if -epsilon < t < epsilon:
                    if direction_sweep_x * direction_static_x + direction_sweep_y * direction_static_y < 0:
                        m_new = m_new - direction_m
                        n_new = n_new - direction_n

                elif (1 - epsilon) < t < (1 + epsilon):
                    if direction_sweep_x * direction_static_x + direction_sweep_y * direction_static_y > 0:
                        m_new = m_new + direction_m
                        n_new = n_new + direction_n

                if (1 - epsilon) < u < (1 + epsilon):
                    pass
                else:
                    j_new = j

                u_min = u

                # plt.scatter(point_sweep_2x, point_sweep_2y)

                # break

        elif math.isnan(u) or math.isnan(t):

            if point_sweep_1x != point_sweep_2x:
                u1 = (vertex_static_1x - point_sweep_1x) / (point_sweep_2x - point_sweep_1x)
                u2 = (vertex_static_2x - point_sweep_1x) / (point_sweep_2x - point_sweep_1x)
            else:
                u1 = (vertex_static_1y - point_sweep_1y) / (point_sweep_2y - point_sweep_1y)
                u2 = (vertex_static_2y - point_sweep_1y) / (point_sweep_2y - point_sweep_1y)

            if u1 > u_min:
                continue

            # print("u1", u1)
            # print("u2", u2)

            if u1 > u2:
                u1, u2 = u2, u1

            if (0 - epsilon) < u1 < (1 + epsilon):

                sweep_is_inside_static = True

                m_new = cells_border_static_m[index_vertex_1]
                n_new = cells_border_static_n[index_vertex_1]

                if (1 - epsilon) < u < (1 + epsilon):
                    pass
                else:
                    j_new = j

                u_min = u1

                # break

    # if not input_is_sweep:
    #     i_left = i - 1
    #     i_right = i
    #     area_sweep = (point_sweep_1x * point_sweep_2y - point_sweep_2x * point_sweep_1y) / 2
    #
    #     if 0 <= i_left < (shape_sweep_x - 1):
    #         area_output[i_left, j] += area_sweep
    #     if 0 <= i_right < (shape_sweep_x - 1):
    #         area_output[i_right, j] -= area_sweep

    if math.isfinite(u_min):
        point_sweep_2x = point_sweep_1x + u_min * (point_sweep_2x - point_sweep_1x)
        point_sweep_2y = point_sweep_1y + u_min * (point_sweep_2y - point_sweep_1y)

    # print("point_sweep_1", (point_sweep_1x, point_sweep_1y))
    # print("point_sweep_2", (point_sweep_2x, point_sweep_2y))

    # plt.scatter(point_sweep_1x, point_sweep_1y)
    # plt.scatter(point_sweep_2x, point_sweep_2y)

    return point_sweep_2x, point_sweep_2y, j_new, m_new, n_new, sweep_is_inside_static


@numba.njit(cache=True, inline="always")
def _step_inside_static(
        # values_input: np.ndarray,
        # values_output: np.ndarray,
        area_input: np.ndarray,
        grid_static_x: np.ndarray,
        grid_static_y: np.ndarray,
        shape_sweep_x: int,
        shape_sweep_y: int,
        point_sweep_1x: float,
        point_sweep_1y: float,
        point_sweep_2x: float,
        point_sweep_2y: float,
        axis: int,
        input_is_sweep: bool,
        epsilon: float,
        m_old: int,
        n_old: int,
        i: int,
        j: int,
        m: int,
        n: int,
        weight: numba.typed.List,
) -> tuple[float, float, int, int, int]:

    # print("step inside")

    # intersection_found = False

    shape_static_x, shape_static_y = grid_static_x.shape

    vertices_static_m = (0, 1, 1, 0)
    vertices_static_n = (0, 0, 1, 1)

    direction_sweep_x = point_sweep_2x - point_sweep_1x
    direction_sweep_y = point_sweep_2y - point_sweep_1y

    # if axis == 0:
    #     normal_sweep_x = -direction_sweep_y
    #     normal_sweep_y = direction_sweep_x
    # else:
    normal_sweep_x = direction_sweep_y
    normal_sweep_y = -direction_sweep_x

    j_new = j + 1
    m_new = m
    n_new = n

    m_left = m_right = m
    n_left = n_right = n

    u_min = np.inf
    min_is_coincident = False

    for v in range(len(vertices_static_m)):

        index_vertex_static_1 = v - 1
        index_vertex_static_2 = v

        vertex_static_1m = m + vertices_static_m[index_vertex_static_1]
        vertex_static_1n = n + vertices_static_n[index_vertex_static_1]

        vertex_static_2m = m + vertices_static_m[index_vertex_static_2]
        vertex_static_2n = n + vertices_static_n[index_vertex_static_2]

        direction_m = vertex_static_2m - vertex_static_1m
        direction_n = vertex_static_2n - vertex_static_1n

        normal_m = direction_n
        normal_n = -direction_m

        vertex_static_1x = grid_static_x[vertex_static_1m, vertex_static_1n]
        vertex_static_1y = grid_static_y[vertex_static_1m, vertex_static_1n]

        vertex_static_2x = grid_static_x[vertex_static_2m, vertex_static_2n]
        vertex_static_2y = grid_static_y[vertex_static_2m, vertex_static_2n]

        direction_static_x = vertex_static_2x - vertex_static_1x
        direction_static_y = vertex_static_2y - vertex_static_1y

        if axis == 0:
            normal_static_x = -direction_static_y
            normal_static_y = direction_static_x
        else:
            normal_static_x = direction_static_y
            normal_static_y = -direction_static_x

        # plt.plot(
        #     [vertex_static_1x, vertex_static_2x],
        #     [vertex_static_1y, vertex_static_2y],
        #     color="lime"
        # )

        t, u = _two_line_segment_intersection_parameters(
            x1=vertex_static_1x, y1=vertex_static_1y,
            x2=vertex_static_2x, y2=vertex_static_2y,
            x3=point_sweep_1x, y3=point_sweep_1y,
            x4=point_sweep_2x, y4=point_sweep_2y,
        )

        # print(t, u)

        if u > u_min:
            # print("u greater than u_min, continuing")
            continue

        if (0 - epsilon) < u < (1 + epsilon):
            if (0 - epsilon) < t < (1 + epsilon):

                if min_is_coincident:
                    continue

                # print("intersection found")

                # if intersection_found:
                #     continue

                m_test = m + normal_m
                n_test = n + normal_n

                if -epsilon < t < epsilon:
                    m_test = m_test - direction_m
                    n_test = n_test - direction_n
                elif (1 - epsilon) < t < (1 + epsilon):
                    m_test = m_test + direction_m
                    n_test = n_test + direction_n

                if (m_test == m_old) and (n_test == n_old):
                    continue

                if -epsilon < u < epsilon:
                    projection_x = (point_sweep_2x - point_sweep_1x) * normal_static_x
                    projection_y = (point_sweep_2y - point_sweep_1y) * normal_static_y
                    projection = projection_x + projection_y
                    # print("projection", projection)
                    if projection < 0:
                        continue
                    else:
                        j_new = j
                elif (1 - epsilon) < u < (1 + epsilon):
                    pass
                else:
                    j_new = j

                u_min = u

                m_new = m_test
                n_new = n_test

                # intersection_found = True

                # plt.scatter(point_sweep_2x, point_sweep_2y)

        elif math.isnan(u) or math.isnan(t):

            # print("parallel lines found")

            if point_sweep_1x != point_sweep_2x:
                u1 = (vertex_static_1x - point_sweep_1x) / (point_sweep_2x - point_sweep_1x)
                u2 = (vertex_static_2x - point_sweep_1x) / (point_sweep_2x - point_sweep_1x)
            else:
                u1 = (vertex_static_1y - point_sweep_1y) / (point_sweep_2y - point_sweep_1y)
                u2 = (vertex_static_2y - point_sweep_1y) / (point_sweep_2y - point_sweep_1y)

            # print("u1", u1)
            # print("u2", u2)

            if u1 > u_min:
                continue

            if u1 > u2:
                u1, u2 = u2, u1
                direction_sweep_m = -direction_m
                direction_sweep_n = -direction_n
            else:
                direction_sweep_m = direction_m
                direction_sweep_n = direction_n

            if u1 < epsilon:
                # print("boop")
                if (0 - epsilon) < u2:
                    if u2 < (1 + epsilon):
                        m_new = m + direction_sweep_m
                        n_new = n + direction_sweep_n
                        if (1 - epsilon) < u2 < (1 + epsilon):
                            pass
                        else:
                            j_new = j

                        u_min = u2
                        min_is_coincident = True
                            # plt.scatter(point_sweep_2x, point_sweep_2y, zorder=10)

                    m_left = m
                    m_right = m + normal_m

                    n_left = n
                    n_right = n + normal_n

                # plt.plot(
                #     [point_sweep_1x, point_sweep_1x + normal_sweep_x],
                #     [point_sweep_1y, point_sweep_1y + normal_sweep_y],
                #     color="purple"
                # )
                # plt.plot(
                #     [vertex_static_1x, vertex_static_1x + normal_static_x],
                #     [vertex_static_1x, vertex_static_1x + normal_static_y],
                #     color="orange"
                # )

                if normal_sweep_x * normal_static_x + normal_sweep_y * normal_static_y < 0:
                    # print("switch")
                    m_left, m_right = m_right, m_left
                    n_left, n_right = n_right, n_left

                # break

            elif epsilon <= u1 <= (1 + epsilon):

                m_new = m + direction_sweep_m
                n_new = n + direction_sweep_n

                if (m_new == m_old) and (n_new == n_old):
                    continue

                if (1 - epsilon) < u1 < (1 + epsilon):
                    pass
                else:
                    j_new = j

                u_min = u1
                min_is_coincident = True
                # break

    if math.isfinite(u_min):
        point_sweep_2x = point_sweep_1x + u_min * (point_sweep_2x - point_sweep_1x)
        point_sweep_2y = point_sweep_1y + u_min * (point_sweep_2y - point_sweep_1y)

    i_left, i_right = i - 1, i

    # print("i_left", i_left)
    # print("i_right", i_right)
    # print("m_left", m_left)
    # print("m_right", m_right)
    # print("n_left", n_left)
    # print("n_right", n_right)

    if input_is_sweep:
        shape_input_x = shape_sweep_x
        shape_input_y = shape_sweep_y
        shape_output_x = shape_static_x
        shape_output_y = shape_static_y

        i_input_left = i_left
        i_input_right = i_right
        j_input_left = j
        j_input_right = j

        i_output_left = m_left
        i_output_right = m_right
        j_output_left = n_left
        j_output_right = n_right

    else:
        shape_input_x = shape_static_x
        shape_input_y = shape_static_y
        shape_output_x = shape_sweep_x
        shape_output_y = shape_sweep_y

        i_input_left = m_left
        i_input_right = m_right
        j_input_left = n_left
        j_input_right = n_right

        i_output_left = i_left
        i_output_right = i_right
        j_output_left = j
        j_output_right = j

    i_input_right = int(i_input_right)
    i_output_right = int(i_output_right)

    if axis == 0:
        if input_is_sweep:
            shape_input_x, shape_input_y = shape_input_y, shape_input_x
            i_input_left, j_input_left = j_input_left, i_input_left
            i_input_right, j_input_right = j_input_right, i_input_right
        else:
            shape_output_x, shape_output_y = shape_output_y, shape_output_x
            i_output_left, j_output_left = j_output_left, i_output_left
            i_output_right, j_output_right = j_output_right, i_output_right

    area_sweep = (point_sweep_1x * point_sweep_2y - point_sweep_2x * point_sweep_1y) / 2

    if 0 <= i_input_left < (shape_input_x - 1) and 0 <= j_input_left < (shape_input_y - 1):
        if 0 <= i_output_left < (shape_output_x - 1) and 0 <= j_output_left < (shape_output_y - 1):
            weight.append((
                (shape_input_y - 1) * i_input_left + j_input_left,
                (shape_output_y - 1) * i_output_left + j_output_left,
                area_sweep / area_input[i_input_left, j_input_left],
            ))

    if 0 <= i_input_right < (shape_input_x - 1) and 0 <= j_input_right < (shape_input_y - 1):
        if 0 <= i_output_right < (shape_output_x - 1) and 0 <= j_output_right < (shape_output_y - 1):
            weight.append((
                (shape_input_y - 1) * i_input_right + j_input_right,
                (shape_output_y - 1) * i_output_right + j_output_right,
                -area_sweep / area_input[i_input_right, j_input_right],
            ))

    # if 0 <= i_input_left < (shape_input_x - 1) and 0 <= j_input_left < (shape_input_y - 1):
    #     value_input_left = values_input[i_input_left, j_input_left]
    # else:
    #     value_input_left = 0
    #
    # if 0 <= i_input_right < (shape_input_x - 1) and 0 <= j_input_right < (shape_input_y - 1):
    #     value_input_right = values_input[int(i_input_right), j_input_right]
    # else:
    #     value_input_right = 0
    #
    # if 0 <= i_output_left < (shape_output_x - 1) and 0 <= j_output_left < (shape_output_y - 1):
    #     values_output[i_output_left, j_output_left] += value_input_left * area_sweep / 2
    #
    # if 0 <= i_output_right < (shape_output_x - 1) and 0 <= j_output_right < (shape_output_y - 1):
    #     values_output[int(i_output_right), j_output_right] -= value_input_right * area_sweep / 2

    # if not input_is_sweep:
    #     area_output[i_output_left, j_output_left] += area_sweep
    #     area_output[int(i_output_right), j_output_right] -= area_sweep

    # print("point_sweep_1", (point_sweep_1x, point_sweep_1y))
    # print("point_sweep_2", (point_sweep_2x, point_sweep_2y))
    # plt.scatter(point_sweep_1x, point_sweep_1y, zorder=10)
    # plt.scatter(point_sweep_2x, point_sweep_2y, zorder=10)

    return point_sweep_2x, point_sweep_2y, j_new, m_new, n_new


@numba.njit(cache=True, inline="always")
def _grid_area_sweep(
        out: np.ndarray,
        grid_x: np.ndarray,
        grid_y: np.ndarray,
        axis: int,
):

    if axis == 0:
        grid_y, grid_x = grid_x.T, grid_y.T
        out = out.T
    elif axis == 1:
        pass
    else:
        print("axis should be zero or one")
    #     raise ValueError(f"The `axis` argument must be 0 or 1, got {axis}")

    shape_output_x, shape_output_y = grid_x.shape

    for i in numba.prange(shape_output_x):
        i = numba.types.int64(i)

        for j in range(shape_output_y - 1):

            index_vertex_1j = j
            index_vertex_2j = j + 1

            vertex_1x = grid_x[i, index_vertex_1j]
            vertex_1y = grid_y[i, index_vertex_1j]

            vertex_2x = grid_x[i, index_vertex_2j]
            vertex_2y = grid_y[i, index_vertex_2j]

            area_sweep = (vertex_1x * vertex_2y - vertex_2x * vertex_1y) / 2
            # if axis == 0:
            #     area_sweep = -area_sweep

            i_left = i - 1
            i_right = i

            if i_left >= 0:
                out[i_left, j] += area_sweep
            if i_right < (shape_output_x - 1):
                out[i_right, j] -= area_sweep


@numba.njit(cache=True, inline="always", error_model="numpy")
def _indices_of_line_brute(
        grid_x: np.ndarray,
        grid_y: np.ndarray,
        point_1x: float,
        point_1y: float,
        point_2x: float,
        point_2y: float,
        epsilon: float,
) -> tuple[int, int]:

    shape_x, shape_y = grid_x.shape

    vertices_x = np.empty(4)
    vertices_y = np.empty(4)

    for i in range(shape_x - 1):
        for j in range(shape_y - 1):

            v1 = i + 0, j + 0
            v2 = i + 1, j + 0
            v3 = i + 1, j + 1
            v4 = i + 0, j + 1

            vertices_x[0] = grid_x[v1]
            vertices_x[1] = grid_x[v2]
            vertices_x[2] = grid_x[v3]
            vertices_x[3] = grid_x[v4]

            vertices_y[0] = grid_y[v1]
            vertices_y[1] = grid_y[v2]
            vertices_y[2] = grid_y[v3]
            vertices_y[3] = grid_y[v4]

            if _point_is_inside_polygon(
                vertices_x=vertices_x,
                vertices_y=vertices_y,
                point_x=point_1x,
                point_y=point_1y,
                epsilon=epsilon,
            ):
                num_intersections = 0

                for v in range(len(vertices_x)):
                    # print("v", v)

                    index_vertex_1 = v - 1
                    index_vertex_2 = v

                    vertex_1x = vertices_x[index_vertex_1]
                    vertex_1y = vertices_y[index_vertex_1]
                    vertex_2x = vertices_x[index_vertex_2]
                    vertex_2y = vertices_y[index_vertex_2]

                    t, u = _two_line_segment_intersection_parameters(
                        x1=vertex_1x, y1=vertex_1y,
                        x2=vertex_2x, y2=vertex_2y,
                        x3=point_1x, y3=point_1y,
                        x4=point_2x, y4=point_2y,
                    )
                    # print(t, u)

                    if (0 - epsilon) < t < (1 + epsilon):
                        if (0 + epsilon) < u:
                            # print("intersection found")

                            num_intersections += 1

                            if -epsilon < t < epsilon:
                                # print("t is near zero")
                                index_vertex_0 = v - 2
                                vertex_0x = vertices_x[index_vertex_0]
                                vertex_0y = vertices_y[index_vertex_0]
                                direction_x = point_2x - point_1x
                                direction_y = point_2y - point_1y
                                normal_x = direction_y
                                normal_y = -direction_x
                                projection_0x = normal_x * (vertex_1x - vertex_0x)
                                projection_0y = normal_y * (vertex_1y - vertex_0y)
                                projection_1x = normal_x * (vertex_2x - vertex_1x)
                                projection_1y = normal_y * (vertex_2y - vertex_1y)
                                projection_0 = projection_0x + projection_0y
                                projection_1 = projection_1x + projection_1y
                                if (-epsilon < projection_0 < epsilon) or (-epsilon < projection_1 < epsilon):
                                    pass
                                elif math.copysign(1, projection_0) == math.copysign(1, projection_1):
                                    # print("projection_0 and projection_1 have the same sign")
                                    num_intersections -= 1

                # print("num_intersections", num_intersections)

                if (num_intersections % 2) == 1:
                    return i, j

    return 0, 0


@numba.njit(cache=True, inline="always", error_model="numpy")
def _point_is_inside_polygon(
        vertices_x: np.ndarray,
        vertices_y: np.ndarray,
        point_x: float,
        point_y: float,
        epsilon: float
) -> bool:

    x3 = point_x
    y3 = point_y
    x4 = x3 + 1
    y4 = y3

    result = False
    for v in range(len(vertices_x)):
        # print("    result", result)
        # print("    v", v)
        y1 = vertices_y[v - 1] - point_y
        y2 = vertices_y[v] - point_y

        # print("    y1", y1)
        # print("    y2", y2)

        y1_and_y2_have_different_signs = math.copysign(1, y1) != math.copysign(1, y2)
        y1_is_zero = -epsilon < y1 < epsilon
        y2_is_zero = -epsilon < y2 < epsilon
        if y1_is_zero and y2_is_zero:
            # print("    both y values are zero")
            x1 = vertices_x[v - 1] - point_x
            x2 = vertices_x[v] - point_x
            if (-epsilon < x1 < epsilon) or (-epsilon < x2 < epsilon):
                pass
            elif math.copysign(1, x1) != math.copysign(1, x2):
                # print("    x1 and x2 have different signs")
                return True
        elif y1_and_y2_have_different_signs or y1_is_zero or y2_is_zero:
            # print("    y values have different sign or degenerate")

            x1 = vertices_x[v - 1] - point_x
            x2 = vertices_x[v] - point_x

            # print("    x1", x1)
            # print("    x2", x2)

            if (x1 > epsilon) and (x2 > epsilon):
                # print("    both x values are larger")
                result = not result
                if y1_is_zero:
                    y0 = vertices_y[v - 2] - point_y
                    if (-epsilon < y0 < epsilon) or (-epsilon < y2 <epsilon):
                        pass
                    elif math.copysign(1, y0) != math.copysign(1, y2):
                        # print("    y0 and y2 have different signs")
                        result = not result
            elif (x1 < -epsilon) and (x2 < -epsilon):
                # print("    both x values are smaller")
                pass
            else:
                # print("    x values have different sign or degenerate")
                t, u = _two_line_segment_intersection_parameters(
                    x1=x1, y1=y1,
                    x2=x2, y2=y2,
                    x3=0, y3=0,
                    x4=1, y4=0,
                )
                # print("    t", t)
                # print("    u", u)

                if (0 - epsilon) <= t < (1 + epsilon):
                    if -epsilon < u < epsilon:
                        # print("    point on border")
                        return True
                    elif u > 0:
                        # print("    intersection detected")
                        result = not result
                        if -epsilon < t < epsilon:
                            # print("     intersection with start point")
                            y0 = vertices_y[v - 2] - point_y
                            if (-epsilon < y0 < epsilon) or (-epsilon < y2 <epsilon):
                                pass
                            elif math.copysign(1, y0) != math.copysign(1, y2):
                                # print("    y0 and y2 have different signs")
                                result = not result


        # slope = (y0 - y1) / (x0 - x1)
        # condition_1 = (y1 >= point_y) != (y0 >= point_y)
        # condition_2 = point_x < ((point_y - y1) / slope + x1)
        # result = result ^ (condition_1 & condition_2)

    return result


@numba.njit(cache=True, inline="always", error_model="numpy")
def _two_line_segment_intersection(
        x1: float, y1: float,
        x2: float, y2: float,
        x3: float, y3: float,
        x4: float, y4: float,
) -> tuple[float, float]:

    t_numerator = (x1 - x3) * (y3 - y4) - (y1 - y3) * (x3 - x4)
    t_denominator = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)

    u_numerator = (x1 - x3) * (y1 - y2) - (y1 - y3) * (x1 - x2)
    u_denominator = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)

    t = t_numerator / t_denominator
    u = u_numerator / u_denominator

    if (0 < t <= 1) and (0 < u <= 1):
        x = x1 + t * (x2 - x1)
        y = y1 + t * (y2 - y1)
        return x, y

    else:
        return np.nan, np.nan


@numba.njit(cache=True, inline="always", error_model="numpy")
def _two_line_segment_intersection_parameters(
        x1: float, y1: float,
        x2: float, y2: float,
        x3: float, y3: float,
        x4: float, y4: float,
) -> tuple[float, float]:

    t_numerator = (x1 - x3) * (y3 - y4) - (y1 - y3) * (x3 - x4)
    t_denominator = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)

    u_numerator = (x1 - x3) * (y1 - y2) - (y1 - y3) * (x1 - x2)
    u_denominator = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)

    # print("t_numerator", t_numerator)
    # print("t_denominator", t_denominator)
    # print("u_numerator", u_numerator)
    # print("u_denominator", u_denominator)

    t = t_numerator / t_denominator
    u = u_numerator / u_denominator

    return t, u



import numpy as np
import numba


def _find_indices_brute(
    coordinates_input: tuple[np.ndarray, ...],
    coordinates_output: tuple[np.ndarray, ...],
    fill_value: None | int = None,
) -> tuple[np.ndarray, ...]:
    if len(coordinates_input) == 1:
        result = _find_indices_brute_1d(
            coordinates_input=coordinates_input,
            coordinates_output=coordinates_output,
            fill_value=fill_value,
        )
    else:
        raise ValueError(
            f"{len(coordinates_input)}-dimensional brute-force search not supported"
        )

    return result


@numba.njit(cache=True, parallel=True)
def _find_indices_brute_1d(
    coordinates_input: tuple[np.ndarray],
    coordinates_output: tuple[np.ndarray],
    fill_value: int,
) -> tuple[np.ndarray]:
    (x_input,) = coordinates_input
    (x_output,) = coordinates_output

    num_d, num_m = x_input.shape
    num_d, num_i = x_output.shape

    result = np.empty(shape=x_output.shape, dtype=np.int64)

    for d in numba.prange(num_d):
        x_input_d = x_input[d]
        x_output_d = x_output[d]

        for i in range(num_i):
            x_output_di = x_output_d[i]
            for m in range(num_m - 1):
                if x_input_d[m] <= x_output_di <= x_input_d[m + 1]:
                    result[d, i] = m
                    break
            else:
                result[d, i] = fill_value

    return (result,)



import numba

__all__ = [
    "sign",
]


@numba.njit(cache=True, inline="always", error_model="numpy")
def sign(x: float):
    """
    Numba-compiled version of the `sign function <https://en.wikipedia.org/wiki/Sign_function>`_

    Parameters
    ----------
    x
        the value to find the sign of
    """
    return bool(x > 0) - bool(x < 0)



from typing import Sequence
import numpy as np
import numba
from regridding import _util

__all__ = [
    "regrid_from_weights",
]


def regrid_from_weights(
    weights: np.ndarray,
    shape_input: tuple[int, ...],
    shape_output: tuple[int, ...],
    values_input: np.ndarray,
    values_output: None | np.ndarray = None,
    axis_input: None | int | Sequence[int] = None,
    axis_output: None | int | Sequence[int] = None,
) -> np.ndarray:
    """
    Regrid an array of values using weights computed by
    :func:`regridding.weights`.

    Parameters
    ----------
    weights
        Ragged array of weights computed by :func:`regridding.weights`.
    shape_input
        Broadcasted shape of the input coordinates computed by :func:`regridding.weights`.
    shape_output
        Broadcasted shape of the output coordinates computed by :func:`regridding.weights`.
    values_input
        Input array of values to be resampled.
    values_output
        Optional array in which to place the output.
    axis_input
        Logical axes of the input array to resample.
        If :obj:`None`, resample all the axes of the input array.
        The number of axes should be equal to the number of
        coordinates in the original input grid passed to :func:`regridding.weights`.
    axis_output
        Logical axes of the output array corresponding to the resampled axes
        of the input array.
        If :obj:`None`, all the axes of the output array correspond to resampled
        axes in the input grid.
        The number of axes should be equal to the original number of
        coordinates in the output grid passed to :func:`regridding.weights`.

    See Also
    --------
    :func:`regridding.regrid`
    :func:`regridding.regrid_from_weights`
    """

    shape_input = np.broadcast_shapes(values_input.shape, shape_input)

    ndim_input = len(shape_input)
    axis_input = _util._normalize_axis(axis_input, ndim=ndim_input)

    shape_orthogonal = (
        1 if i in axis_input else shape_input[i] for i in range(-len(shape_input), 0)
    )
    weights = np.broadcast_to(np.array(weights), shape_orthogonal, subok=True)
    values_input = np.broadcast_to(values_input, shape_input, subok=True)

    if values_output is None:
        shape_output = np.broadcast_shapes(
            shape_output,
            tuple(
                shape_input[ax] if ax not in axis_input else 1
                for ax in _util._normalize_axis(None, ndim_input)
            ),
        )
        values_output = np.zeros_like(values_input, shape=shape_output)
    else:
        if values_output.shape != shape_output:
            raise ValueError(f"")
        values_output.fill(0)

    ndim_output = len(shape_output)
    axis_output = _util._normalize_axis(axis_output, ndim=ndim_output)

    axis_input_numba = ~np.arange(len(axis_input))[::-1]
    axis_output_numba = ~np.arange(len(axis_output))[::-1]

    shape_input_numba = tuple(shape_input[ax] for ax in axis_input)
    shape_output_numba = tuple(shape_output[ax] for ax in axis_output)

    values_input = np.moveaxis(values_input, axis_input, axis_input_numba)
    values_output = np.moveaxis(values_output, axis_output, axis_output_numba)

    shape_output_tmp = values_output.shape

    values_input = values_input.reshape(-1, *shape_input_numba)
    values_output = values_output.reshape(-1, *shape_output_numba)

    weights = numba.typed.List(weights.reshape(-1))

    values_input = np.ascontiguousarray(values_input)
    values_output = np.ascontiguousarray(values_output)

    _regrid_from_weights(
        weights=weights,
        values_input=values_input,
        values_output=values_output,
    )

    values_output = values_output.reshape(*shape_output_tmp)

    values_output = np.moveaxis(values_output, axis_output_numba, axis_output)

    return values_output


@numba.njit(cache=True, parallel=True)
def _regrid_from_weights(
    weights: numba.typed.List,
    values_input: np.ndarray,
    values_output: np.ndarray,
) -> None:

    for d in numba.prange(len(weights)):
        d = numba.types.int64(d)
        weights_d = weights[d]
        values_input_d = values_input[d].reshape(-1)
        values_output_d = values_output[d].reshape(-1)
        for w in range(len(weights_d)):
            i_input, i_output, weight = weights_d[w]
            values_output_d[int(i_output)] += weight * values_input_d[int(i_input)]



from typing import Sequence, Literal
import numpy as np
import regridding
from . import regrid_from_weights

__all__ = [
    "regrid",
]


def regrid(
    coordinates_input: tuple[np.ndarray, ...],
    coordinates_output: tuple[np.ndarray, ...],
    values_input: np.ndarray,
    values_output: None | np.ndarray = None,
    axis_input: None | int | Sequence[int] = None,
    axis_output: None | int | Sequence[int] = None,
    method: Literal["multilinear", "conservative"] = "multilinear",
) -> np.ndarray:
    """
    Regrid an array of values defined on a logically-rectangular curvilinear
    grid onto a new logically-rectangular curvilinear grid.

    Parameters
    ----------
    coordinates_input
        Coordinates of the input grid.
    coordinates_output
        Coordinates of the output grid.
        Should have the same number of coordinates as the input grid.
    values_input
        Input array of values to be resampled.
    values_output
        Optional array in which to place the output.
    axis_input
        Logical axes of the input grid to resample.
        If :obj:`None`, resample all the axes of the input grid.
        The number of axes should be equal to the number of
        coordinates in the input grid.
    axis_output
        Logical axes of the output grid corresponding to the resampled axes
        of the input grid.
        If :obj:`None`, all the axes of the output grid correspond to resampled
        axes in the input grid.
        The number of axes should be equal to the number of
        coordinates in the output grid.
    method
        The type of regridding to use.
        The ``conservative`` method uses the algorithm described in
        :footcite:t:`Ramshaw1985`

    See Also
    --------
    :func:`regridding.regrid`
    :func:`regridding.regrid_from_weights`

    Examples
    --------

    Regrid a 1D array using multilinear interpolation.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import regridding

        # Define the input grid
        x_input = np.linspace(-1, 1, num=11)

        # Define the input array
        values_input = np.square(x_input)

        # Define the output grid
        x_output = np.linspace(-1, 1, num=51)

        # Regrid the input array onto the output grid
        values_output = regridding.regrid(
            coordinates_input=(x_input,),
            coordinates_output=(x_output,),
            values_input=values_input,
            method="multilinear",
        )

        # Plot the results
        plt.figure(figsize=(6, 3));
        plt.scatter(x_input, values_input, s=100, label="input", zorder=1);
        plt.scatter(x_output, values_output, label="interpolated", zorder=0);
        plt.legend();

    |

    Regrid a 2D array using conservative resampling.

    .. jupyter-execute::

        # Define the number of edges in the input grid
        num_x = 66
        num_y = 66

        # Define a dummy linear grid
        x = np.linspace(-5, 5, num=num_x)
        y = np.linspace(-5, 5, num=num_y)
        x, y = np.meshgrid(x, y, indexing="ij")

        # Define the curvilinear input grid using the dummy grid
        angle = 0.4
        x_input = x * np.cos(angle) - y * np.sin(angle) + 0.05 * x * x
        y_input = x * np.sin(angle) + y * np.cos(angle) + 0.05 * y * y

        # Define the test pattern
        pitch = 16
        a_input = 0 * x[:~0,:~0]
        a_input[::pitch, :] = 1
        a_input[:, ::pitch] = 1
        a_input[pitch//2::pitch, pitch//2::pitch] = 1

        # Define a rectilinear output grid using the limits of the input grid
        x_output = np.linspace(x_input.min(), x_input.max(), num_x // 2)
        y_output = np.linspace(y_input.min(), y_input.max(), num_y // 2)
        x_output, y_output = np.meshgrid(x_output, y_output, indexing="ij")

        # Regrid the test pattern onto the new grid
        a_output = regridding.regrid(
            coordinates_input=(x_input, y_input),
            coordinates_output=(x_output, y_output),
            values_input=a_input,
            method="conservative",
        )

        fig, axs = plt.subplots(
            ncols=2,
            sharex=True,
            sharey=True,
            figsize=(8, 4),
            constrained_layout=True,
        );
        axs[0].pcolormesh(x_input, y_input, a_input);
        axs[0].set_title("input array");
        axs[1].pcolormesh(x_output, y_output, a_output);
        axs[1].set_title("regridded array");

    |

    References
    ----------
    .. footbibliography::
    """
    weights, shape_input, shape_output = regridding.weights(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        axis_input=axis_input,
        axis_output=axis_output,
        method=method,
    )
    result = regrid_from_weights(
        weights=weights,
        shape_input=shape_input,
        shape_output=shape_output,
        values_input=values_input,
        values_output=values_output,
        axis_input=axis_input,
        axis_output=axis_output,
    )
    return result



import pytest
import numpy as np
import regridding

x = np.linspace(-1, 1, num=10)
y = np.linspace(-1, 1, num=11)
x_broadcasted, y_broadcasted = np.meshgrid(
    x,
    y,
    indexing="ij",
)

new_y = np.linspace(-1, 1, num=5)
new_x = np.linspace(-1, 1, num=6)

new_x_broadcasted, new_y_broadcasted = np.meshgrid(
    x,
    new_y,
    indexing="ij",
)

new_x_broadcasted_2, new_y_broadcasted_2 = np.meshgrid(
    new_x,
    y,
    indexing="ij",
)


@pytest.mark.parametrize(
    argnames="coordinates_input,coordinates_output,values_input,values_output,axis_input,axis_output,result_expected",
    argvalues=[
        (
            (np.linspace(-1, 1, num=11),),
            (np.linspace(-1, 1, num=11),),
            np.square(np.linspace(-1, 1, num=11)),
            None,
            None,
            None,
            np.square(np.linspace(-1, 1, num=11)),
        ),
        (
            (np.linspace(-1, 1, num=11),),
            (np.linspace(-1, 1, num=11),),
            np.square(np.linspace(-1, 1, num=11)),
            np.empty(shape=11),
            None,
            None,
            np.square(np.linspace(-1, 1, num=11)),
        ),
        (
            (y,),
            (new_y,),
            x_broadcasted + y_broadcasted,
            None,
            (~0,),
            (~0,),
            new_x_broadcasted + new_y_broadcasted,
        ),
        (
            (x[..., np.newaxis],),
            (new_x[..., np.newaxis],),
            x_broadcasted + y_broadcasted,
            None,
            (0,),
            (0,),
            new_x_broadcasted_2 + new_y_broadcasted_2,
        ),
        (
            (x[..., np.newaxis],),
            (0.1 * new_x[..., np.newaxis] + 0.001 * new_y,),
            x[..., np.newaxis],
            None,
            (0,),
            (0,),
            0.1 * new_x[..., np.newaxis] + 0.001 * new_y,
        ),
    ],
)
def test_regrid_multilinear_1d(
    coordinates_input: tuple[np.ndarray, ...],
    coordinates_output: tuple[np.ndarray, ...],
    values_input: np.ndarray,
    values_output: None | np.ndarray,
    axis_input: None | int | tuple[int, ...],
    axis_output: None | int | tuple[int, ...],
    result_expected: np.ndarray,
):
    result = regridding.regrid(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        values_input=values_input,
        values_output=values_output,
        axis_input=axis_input,
        axis_output=axis_output,
        method="multilinear",
    )
    assert isinstance(result, np.ndarray)
    assert np.issubdtype(result.dtype, float)
    assert np.allclose(result, result_expected)


@pytest.mark.parametrize(
    argnames="coordinates_input, values_input, axis_input, coordinates_output, values_output, axis_output",
    argvalues=[
        (
            (x_broadcasted, y_broadcasted),
            np.random.normal(size=(10 - 1, 11 - 1)),
            None,
            (1.1 * x_broadcasted + 0.01, 1.2 * y_broadcasted + 0.01),
            None,
            None,
        ),
        (
            (
                x_broadcasted[..., np.newaxis] + np.array([0, 0.001]),
                y_broadcasted[..., np.newaxis] + np.array([0, 0.001]),
            ),
            np.random.normal(size=(x.shape[0] - 1, y.shape[0] - 1, 2)),
            (0, 1),
            (
                1.1 * (x_broadcasted[..., np.newaxis] + np.array([0, 0.001])) + 0.01,
                1.2 * (y_broadcasted[..., np.newaxis] + np.array([0, 0.01])) + 0.001,
            ),
            None,
            (0, 1),
        ),
    ],
)
def test_regrid_conservative_2d(
    coordinates_input: tuple[np.ndarray, ...],
    coordinates_output: tuple[np.ndarray, ...],
    values_input: np.ndarray,
    values_output: None | np.ndarray,
    axis_input: None | int | tuple[int, ...],
    axis_output: None | int | tuple[int, ...],
):
    result = regridding.regrid(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        values_input=values_input,
        values_output=values_output,
        axis_input=axis_input,
        axis_output=axis_output,
        method="conservative",
    )

    result_shape = np.array(np.broadcast(*coordinates_output).shape)

    if axis_input is None:
        result_shape = result_shape - 1
    else:
        for ax in axis_input:
            result_shape[ax] = result_shape[ax] - 1

    assert np.issubdtype(result.dtype, float)
    assert result.shape == tuple(result_shape)
    assert np.isclose(result.sum(), values_input.sum())



"""
Numba-compiled computational geometry routines needed for regridding operations.
"""

import math
import numpy as np
import numba
import regridding

__all__ = [
    "line_equation_2d",
    "bounding_boxes_intersect_2d",
    "two_line_segment_intersection_parameters",
    "point_is_inside_polygon",
]


@numba.njit(cache=True, inline="always", error_model="numpy")
def line_equation_2d(
    x: float,
    y: float,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
) -> float:
    """
    Test if a given point lies above, on, or below a line specified by two
    endpoints.

    Returns zero if a point is on the line.
    Otherwise, points above the line have the opposite sign of points below
    the line.

    Parameters
    ----------
    x
        :math:`x`-component of the test point
    y
        :math:`y`-component of the test point
    x1
        :math:`x`-component of the first endpoint of the line
    y1
        :math:`y`-component of the first endpoint of the line
    x2
        :math:`x`-component of the second endpoint of the line
    y2
        :math:`y`-component of the second endpoint of the line
    """
    result = (x - x1) * (y2 - y1) - (y - y1) * (x2 - x1)
    return result


@numba.njit(cache=True, inline="always", error_model="numpy")
def bounding_boxes_intersect_2d(
    x_p1: float,
    y_p1: float,
    x_p2: float,
    y_p2: float,
    x_q1: float,
    y_q1: float,
    x_q2: float,
    y_q2: float,
) -> bool:
    """
    Test if two bounding boxes, :math:`p` and :math:`q`, intersect.

    If the edges of the two boxes are touching, it's counted as an intersection.

    Parameters
    ----------
    x_p1
        :math:`x`-coordinate of the first point of box :math:`p`
    y_p1
        :math:`y`-coordinate of the first point of box :math:`p`
    x_p2
        :math:`x`-coordinate of the second point of box :math:`p`
    y_p2
        :math:`y`-coordinate of the second point of box :math:`p`
    x_q1
        :math:`x`-coordinate of the first point of box :math:`q`
    y_q1
        :math:`y`-coordinate of the first point of box :math:`q`
    x_q2
        :math:`x`-coordinate of the second point of box :math:`q`
    y_q2
        :math:`y`-coordinate of the second point of box :math:`q`

    Examples
    --------

    Create 4 boxes :math:`P`, :math:`Q`, :math:`R`, and :math:`S`.
    Check if boxes :math:`Q`, :math:`R`, and :math:`S` intersect with
    box :math:`P`, and plot the boxes as filled if they intersect,
    and unfilled if they don't intersect.

    .. jupyter-execute::

        import matplotlib as mpl
        import matplotlib.pyplot as plt
        import regridding

        # box P
        x_p1 = 0
        y_p1 = 0
        x_p2 = 1
        y_p2 = 1

        # box Q
        x_q1 = 1.1
        y_q1 = 1.1
        x_q2 = 2.1
        y_q2 = 2.1

        # box R
        x_r1 = 1
        y_r1 = 1
        x_r2 = 2
        y_r2 = 2

        # box S
        x_s1 = 0.9
        y_s1 = 0.9
        x_s2 = 1.9
        y_s2 = 1.9

        p_and_q_intersect = regridding.geometry.bounding_boxes_intersect_2d(
            x_p1=x_p1, y_p1=y_p1,
            x_p2=x_p2, y_p2=y_p2,
            x_q1=x_q1, y_q1=y_q1,
            x_q2=x_q2, y_q2=y_q2,
        )

        p_and_r_intersect = regridding.geometry.bounding_boxes_intersect_2d(
            x_p1=x_p1, y_p1=y_p1,
            x_p2=x_p2, y_p2=y_p2,
            x_q1=x_r1, y_q1=y_r1,
            x_q2=x_r2, y_q2=y_r2,
        )

        p_and_s_intersect = regridding.geometry.bounding_boxes_intersect_2d(
            x_p1=x_p1, y_p1=y_p1,
            x_p2=x_p2, y_p2=y_p2,
            x_q1=x_s1, y_q1=y_s1,
            x_q2=x_s2, y_q2=y_s2,
        )

        fig, axs = plt.subplots(ncols=3, sharex=True, sharey=True, figsize=(9, 3), constrained_layout=True)
        axs[0].add_patch(
            mpl.patches.Rectangle(xy=(x_p1, y_p1), width=x_p2 - x_p1, height=y_p2 - y_p1, fill=p_and_q_intersect)
        )
        axs[0].add_patch(
            mpl.patches.Rectangle(xy=(x_q1, y_q1), width=x_q2 - x_q1, height=y_q2 - y_q1, fill=p_and_q_intersect)
        )
        axs[1].add_patch(
            mpl.patches.Rectangle(xy=(x_p1, y_p1), width=x_p2 - x_p1, height=y_p2 - y_p1, fill=p_and_r_intersect)
        )
        axs[1].add_patch(
            mpl.patches.Rectangle(xy=(x_r1, y_r1), width=x_r2 - x_r1, height=y_r2 - y_r1, fill=p_and_r_intersect)
        )
        axs[2].add_patch(
            mpl.patches.Rectangle(xy=(x_p1, y_p1), width=x_p2 - x_p1, height=y_p2 - y_p1, fill=p_and_s_intersect)
        )
        axs[2].add_patch(
            mpl.patches.Rectangle(xy=(x_s1, y_s1), width=x_s2 - x_s1, height=y_s2 - y_s1, fill=p_and_s_intersect)
        )
        axs[0].autoscale_view();
    """
    if x_p1 > x_p2:
        x_p1, x_p2 = x_p2, x_p1
    if y_p1 > y_p2:
        y_p1, y_p2 = y_p2, y_p1
    if x_q1 > x_q2:
        x_q1, x_q2 = x_q2, x_q1
    if y_q1 > y_q2:
        y_q1, y_q2 = y_q2, y_q1

    dx_p = x_p2 - x_p1
    dy_p = y_p2 - y_p1
    dx_q = x_q2 - x_q1
    dy_q = y_q2 - y_q1

    x_p = x_p1 + dx_p / 2
    y_p = y_p1 + dy_p / 2
    x_q = x_q1 + dx_q / 2
    y_q = y_q1 + dy_q / 2

    intersect_x = (2 * abs(x_p - x_q)) <= (dx_p + dx_q)
    intersect_y = (2 * abs(y_p - y_q)) <= (dy_p + dy_q)

    return intersect_x and intersect_y


@numba.njit(cache=True, inline="always", error_model="numpy")
def two_line_segment_intersection_parameters(
    x_p1: float,
    y_p1: float,
    x_p2: float,
    y_p2: float,
    x_q1: float,
    y_q1: float,
    x_q2: float,
    y_q2: float,
) -> tuple[float, float, float]:
    r"""
    Computes the parameters
    (:math:`\text{sdet}`, :math:`\text{tdet}`, and :math:`\text{det}`)
    associated with the intersection of two 2D line segments,
    :math:`p` and :math:`q`.

    This function uses the method described in :footcite:t:`GraphicsGemsIII`.
    :math:`\text{sdet}`, :math:`\text{tdet}`, and :math:`\text{det}` are expected
    to be used to compute the quantities

    .. math::

        s &= \text{sdet} / \text{det} \\
        t &= \text{tdet} / \text{det},

    which can be used to compute the intersection :math:`(x, y)` using the
    parametric equations of the lines:

    .. math::

        x &= (1 - s) * x_{\text{q1}} + s * x_{\text{q2}} = (1 - t) * x_{\text{p1}} + t * x_{\text{p2}} \\
        y &= (1 - s) * y_{\text{q1}} + s * y_{\text{q2}} = (1 - t) * y_{\text{p1}} + t * y_{\text{p2}}.

    The quantities :math:`s` and :math:`t` are not computed directly
    because doing so leads to a loss of precision.

    Parameters
    ----------
    x_p1
        :math:`x` component of the first point in line :math:`p`
    y_p1
        :math:`y` component of the first point in line :math:`p`
    x_p2
        :math:`x` component of the second point in line :math:`p`
    y_p2
        :math:`y` component of the second point in line :math:`p`
    x_q1
        :math:`x` component of the first point in line :math:`q`
    y_q1
        :math:`y` component of the first point in line :math:`q`
    x_q2
        :math:`x` component of the second point in line :math:`q`
    y_q2
        :math:`x` component of the second point in line :math:`q`

    Examples
    --------

    Plot two lines and their intersection

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import regridding

        # line P
        x_p1 = 0
        y_p1 = 0
        x_p2 = 2
        y_p2 = 2

        # line Q
        x_q1 = 0
        y_q1 = 1
        x_q2 = 2
        y_q2 = 0

        sdet, tdet, det = regridding.geometry.two_line_segment_intersection_parameters(
            x_p1=x_p1, y_p1=y_p1,
            x_p2=x_p2, y_p2=y_p2,
            x_q1=x_q1, y_q1=y_q1,
            x_q2=x_q2, y_q2=y_q2,
        )

        s = sdet / det

        x = (1 - s) * x_p1 + s * x_p2
        y = (1 - s) * y_p1 + s * y_p2

        plt.figure()
        plt.plot([x_p1, x_p2], [y_p1, y_p2], label="line $p$")
        plt.plot([x_q1, x_q2], [y_q1, y_q2], label="line $q$")
        plt.scatter(x, y, color="black", zorder=10, label="intersection")
        plt.legend();

    |

    References
    ----------

    .. footbibliography::

    """
    bounding_boxes_intersect = bounding_boxes_intersect_2d(
        x_p1=x_p1,
        y_p1=y_p1,
        x_p2=x_p2,
        y_p2=y_p2,
        x_q1=x_q1,
        y_q1=y_q1,
        x_q2=x_q2,
        y_q2=y_q2,
    )
    if not bounding_boxes_intersect:
        return math.inf, math.inf, 1

    a = line_equation_2d(
        x=x_q1,
        y=y_q1,
        x1=x_p1,
        y1=y_p1,
        x2=x_p2,
        y2=y_p2,
    )
    b = line_equation_2d(
        x=x_q2,
        y=y_q2,
        x1=x_p1,
        y1=y_p1,
        x2=x_p2,
        y2=y_p2,
    )
    if (a != 0) and (b != 0):
        if regridding.math.sign(a) == regridding.math.sign(b):
            return math.inf, math.inf, 1

    c = line_equation_2d(
        x=x_p1,
        y=y_p1,
        x1=x_q1,
        y1=y_q1,
        x2=x_q2,
        y2=y_q2,
    )
    d = line_equation_2d(
        x=x_p2,
        y=y_p2,
        x1=x_q1,
        y1=y_q1,
        x2=x_q2,
        y2=y_q2,
    )
    if (c != 0) and (d != 0):
        if regridding.math.sign(c) == regridding.math.sign(d):
            return math.inf, math.inf, 1

    det = a - b
    sdet = +a
    tdet = -c

    return sdet, tdet, det


@numba.njit(cache=True, inline="always", error_model="numpy")
def point_is_inside_polygon(
    x: float,
    y: float,
    vertices_x: np.ndarray,
    vertices_y: np.ndarray,
) -> bool:
    """
    Check if a given point is inside or on the boundary of a polygon specified
    by its vertices.

    This function uses the extended winding number algorithm described by
    :footcite:t:`Kumar2018`, which addresses boundary issues with the "classic"
    winding number algorithm in :footcite:t:`Alciatore1995`.

    Parameters
    ----------
    x
        :math:`x` component of the test point
    y
        :math:`y` component of the test point
    vertices_x
        :math:`x` component of the polygon's vertices
    vertices_y
        :math:`y` component of the polygon's vertices

    References
    ----------

    .. footbibliography::

    """

    vertices_x = vertices_x - x
    vertices_y = vertices_y - y

    w = 0

    for v in range(len(vertices_x)):
        i = v - 1

        x0 = vertices_x[i + 0]
        y0 = vertices_y[i + 0]
        x1 = vertices_x[i + 1]
        y1 = vertices_y[i + 1]

        if regridding.math.sign(y0) != regridding.math.sign(y1):
            x_intercept = x0 + y0 * (x1 - x0) / (y1 - y0)

            if x_intercept > 0:
                if (y0 == 0) and (x0 > 0):
                    if y1 > 0:
                        w = w + 0.5
                    elif y1 < 0:
                        w = w - 0.5
                elif (y1 == 0) and (x1 > 0):
                    if y0 < 0:
                        w = w + 0.5
                    elif y0 > 0:
                        w = w - 0.5
                elif y0 < 0:
                    w = w + 1
                elif y0 > 0:
                    w = w - 1

            elif x_intercept < 0:
                if (y0 == 0) and (x0 < 0):
                    if y1 < 0:
                        w = w + 0.5
                    elif y1 > 0:
                        w = w - 0.5
                elif (y1 == 0) and (x1 < 0):
                    if y0 > 0:
                        w = w + 0.5
                    elif y0 < 0:
                        w = w - 0.5
                elif y0 > 0:
                    w = w + 1
                elif y0 < 0:
                    w = w - 1

    if w == 0:
        return False
    else:
        return True



from typing import Literal
import pytest
import numpy as np
import regridding


@pytest.mark.parametrize(
    argnames="coordinates_input,coordinates_output",
    argvalues=[
        (
            (np.linspace(-1, 1, num=32),),
            (np.linspace(-1, 1, num=64),),
        ),
        (
            (np.linspace(-1, 1, num=32),),
            (np.linspace(-2, 2, num=64),),
        ),
    ],
)
@pytest.mark.parametrize(
    argnames="method",
    argvalues=[
        "brute",
        "searchsorted",
        pytest.param("invalid method", marks=pytest.mark.xfail),
    ],
)
def test_find_indices_1d(
    coordinates_input: tuple[np.ndarray],
    coordinates_output: tuple[np.ndarray],
    method: Literal["brute", "searchsorted"],
):
    result = regridding.find_indices(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        method=method,
    )

    (coordinates_input_x,) = coordinates_input
    (coordinates_output_x,) = coordinates_output
    (result_x,) = result

    where = coordinates_output_x <= coordinates_input_x.max()
    where &= coordinates_output_x > coordinates_input_x.min()

    assert np.all(
        coordinates_input_x[result_x[where] + 0] <= coordinates_output_x[where]
    )
    assert np.all(
        coordinates_input_x[result_x[where] + 1] >= coordinates_output_x[where]
    )



import pytest
import math
import numpy as np
import regridding


@pytest.mark.parametrize("x", [0, 0.5, 1])
@pytest.mark.parametrize("y_offset", [-1, 0, 2])
@pytest.mark.parametrize("x1", [-1, 0])
@pytest.mark.parametrize("y1", [-2, 0])
@pytest.mark.parametrize("x2", [2, 3])
@pytest.mark.parametrize("y2", [2])
def test_line_equation_2d(
    x: float,
    y_offset: float,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
):
    y_below = (x - x1) * (y2 - y1) / (x2 - x1) + y1 - 1
    y_equal = (x - x1) * (y2 - y1) / (x2 - x1) + y1 + 0
    y_above = (x - x1) * (y2 - y1) / (x2 - x1) + y1 + 1

    result_below = regridding.geometry.line_equation_2d(
        x=x,
        y=y_below,
        x1=x1,
        y1=y1,
        x2=x2,
        y2=y2,
    )
    result_equal = regridding.geometry.line_equation_2d(
        x=x,
        y=y_equal,
        x1=x1,
        y1=y1,
        x2=x2,
        y2=y2,
    )
    result_above = regridding.geometry.line_equation_2d(
        x=x,
        y=y_above,
        x1=x1,
        y1=y1,
        x2=x2,
        y2=y2,
    )

    assert np.isclose(result_equal, 0)
    assert np.sign(result_above) != np.sign(result_below)


@pytest.mark.parametrize(
    argnames="x_p1,y_p1,x_p2,y_p2,x_q1,y_q1,x_q2,y_q2,result_expected",
    argvalues=[
        (0, 0, 1, 1, 0, 0, 1, 1, True),
        (0, 0, 1, 1, 2, 0, 3, 1, False),
        (0, 0, 2, 2, 1, 1, 3, 3, True),
        (0, 0, 1, 1, 1, 1, 2, 2, True),
        (-1, -1, 0, 0, 0, 0, 1, 1, True),
        (-1, 1, 0, 0, 0, 0, 1, 1, True),
        (-1, -1, 0, 0, 1, 0, 0, 1, True),
        (1, 1, 0, 0, 2, 2, 1, 1, True),
    ],
)
def test_bounding_boxes_intersect_2d(
    x_p1: float,
    y_p1: float,
    x_p2: float,
    y_p2: float,
    x_q1: float,
    y_q1: float,
    x_q2: float,
    y_q2: float,
    result_expected: bool,
):
    result = regridding.geometry.bounding_boxes_intersect_2d(
        x_p1=x_p1,
        y_p1=y_p1,
        x_p2=x_p2,
        y_p2=y_p2,
        x_q1=x_q1,
        y_q1=y_q1,
        x_q2=x_q2,
        y_q2=y_q2,
    )
    assert result == result_expected


@pytest.mark.parametrize(
    argnames="x_p1,y_p1,x_p2,y_p2,x_q1,y_q1,x_q2,y_q2,sdet_expected,tdet_expected,det_expected",
    argvalues=[
        (0, 0, 1, 1, 2, 0, 3, 1, math.inf, math.inf, 1),
        (-1, -1, 0, 0, 1, 0, 0, 1, math.inf, math.inf, 1),
        (1, 0, 0, 1, -1, -1, 0, 0, math.inf, math.inf, 1),
        (-1, -1, 1, 1, 1, -1, -1, 1, 4, 4, 8),
        (0, 0, 0, 1, 0, 0, 1, 0, 0, 0, -1),
        (0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0),
        (0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0),
        (0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0),
        (0, 0, 2, 2, 1, 1, 3, 3, 0, 0, 0),
    ],
)
def test_two_line_segment_intersection_parameters(
    x_p1: float,
    y_p1: float,
    x_p2: float,
    y_p2: float,
    x_q1: float,
    y_q1: float,
    x_q2: float,
    y_q2: float,
    sdet_expected: float,
    tdet_expected: float,
    det_expected: float,
):
    sdet, tdet, det = regridding.geometry.two_line_segment_intersection_parameters(
        x_p1=x_p1,
        y_p1=y_p1,
        x_p2=x_p2,
        y_p2=y_p2,
        x_q1=x_q1,
        y_q1=y_q1,
        x_q2=x_q2,
        y_q2=y_q2,
    )

    if math.isnan(sdet_expected):
        assert math.isnan(sdet)
    else:
        assert sdet == sdet_expected

    if math.isnan(tdet_expected):
        assert math.isnan(tdet)
    else:
        assert tdet == tdet_expected

    if math.isnan(det_expected):
        assert math.isnan(det)
    else:
        assert det == det_expected


@pytest.mark.parametrize(
    argnames="x,y,result_expected",
    argvalues=[
        (8, 4, False),
        (7, 4, False),
        (6, 4, False),
        (5, 4, True),
        (4, 4, True),
        (3.5, 4, True),
        (3, 4, True),
        (2, 4, True),
        (1, 4, False),
        (0, 4, False),
        (-1, 4, False),
        (-2, 4, False),
        (-3, 4, False),
        (-4, 4, False),
        (-5, 4, False),
        (-6, 4, False),
        (-7, 4, False),
        (-8, 4, False),
        (3, 8, False),
        (3, 7, False),
        (3, 6, False),
        (3, 5, True),
        (3, 4, True),
        (3, 3.5, True),
        (3, 3, True),
        (3, 2, True),
        (3, 1, False),
        (3, 0, False),
        (3, -1, False),
        (3, -2, False),
        (3, -3, False),
        (3, -4, False),
        (3, -5, False),
        (3, -6, False),
        (3, -7, False),
        (3, -8, False),
    ],
)
def test_point_is_inside_red_cross(
    x: float,
    y: float,
    result_expected: bool,
):
    vertices_x = np.array([2, 2, 3, 3, 4, 4, 3, 3, 2, 2, 1, 1]) + 1
    vertices_y = np.array([2, 1, 1, 2, 2, 3, 3, 4, 4, 3, 3, 2]) + 1

    result = regridding.geometry.point_is_inside_polygon(
        x=x,
        y=y,
        vertices_x=vertices_x,
        vertices_y=vertices_y,
    )
    assert result == result_expected

    result_reversed = regridding.geometry.point_is_inside_polygon(
        x=x,
        y=y,
        vertices_x=vertices_x[::-1],
        vertices_y=vertices_y[::-1],
    )
    assert result_reversed == result_expected



from typing import Literal
import numpy as np
from regridding import _util
from ._find_indices_brute import _find_indices_brute
from ._find_indices_searchsorted import _find_indices_searchsorted

__all__ = [
    "find_indices",
]


def find_indices(
    coordinates_input: tuple[np.ndarray, ...],
    coordinates_output: tuple[np.ndarray, ...],
    axis_input: None | int | tuple[int, ...] = None,
    axis_output: None | int | tuple[int, ...] = None,
    fill_value: None | int = None,
    method: Literal["brute", "searchsorted"] = "brute",
) -> tuple[np.ndarray, ...]:
    """
    Find the index of the input cell which contains the output vertex.

    Parameters
    ----------
    coordinates_input
        the source grid
    coordinates_output
        the destination grid
    axis_input
        the axes in the source grid to search
    axis_output
        the axes in the destination grid corresponding to the source grid
    fill_value
        numeric value to use for invalid indices
    method
        flag to select which search algorithm to use
    """

    (
        coordinates_input,
        coordinates_output,
        axis_input,
        axis_output,
        shape_input,
        shape_output,
        shape_orthogonal,
    ) = _util._normalize_input_output_coordinates(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        axis_input=axis_input,
        axis_output=axis_output,
    )

    if fill_value is None:
        fill_value = np.iinfo(int).max

    axis_input_numba = ~np.arange(len(axis_input))[::-1]
    axis_output_numba = ~np.arange(len(axis_output))[::-1]

    shape_input_numba = tuple(shape_input[ax] for ax in axis_input)
    shape_output_numba = tuple(shape_output[ax] for ax in axis_output)

    coordinates_input = tuple(
        np.moveaxis(v, axis_input, axis_input_numba).reshape(-1, *shape_input_numba)
        for v in coordinates_input
    )
    coordinates_output = tuple(
        np.moveaxis(v, axis_output, axis_output_numba).reshape(-1, *shape_output_numba)
        for v in coordinates_output
    )

    if method == "brute":
        indices_output = _find_indices_brute(
            coordinates_input=coordinates_input,
            coordinates_output=coordinates_output,
            fill_value=fill_value,
        )
    elif method == "searchsorted":
        indices_output = _find_indices_searchsorted(
            coordinates_input=coordinates_input,
            coordinates_output=coordinates_output,
            fill_value=fill_value,
        )
    else:
        raise ValueError(f"method `{method}` not recognized.")

    indices_output = tuple(
        np.moveaxis(
            a=i.reshape(*shape_orthogonal, *shape_output_numba),
            source=axis_output_numba,
            destination=axis_output,
        )
        for i in indices_output
    )

    return indices_output



from typing import Sequence
import numpy as np
import numba
import regridding._util

__all__ = [
    "fill_gauss_seidel",
]


def fill_gauss_seidel(
    a: np.ndarray,
    where: np.ndarray,
    axis: None | int | Sequence[int],
    num_iterations: int = 100,
) -> np.ndarray:

    a = a.copy()

    a, where = np.broadcast_arrays(a, where, subok=True)

    a[where] = 0

    axis = regridding._util._normalize_axis(axis=axis, ndim=a.ndim)
    axis_numba = ~np.arange(len(axis))[::-1]

    shape = a.shape
    shape_numba = tuple(shape[ax] for ax in axis)

    a = np.moveaxis(a, axis, axis_numba)
    where = np.moveaxis(where, axis, axis_numba)

    shape_moved = a.shape

    a = a.reshape(-1, *shape_numba)
    where = where.reshape(-1, *shape_numba)

    if len(axis) == 2:
        result = _fill_gauss_seidel_2d(
            a=a,
            where=where,
            num_iterations=num_iterations,
        )
    else:  # pragma: nocover
        raise ValueError(
            f"The number of interpolation axes, {len(axis)}," f"is not supported"
        )

    result = result.reshape(shape_moved)
    result = np.moveaxis(result, axis_numba, axis)

    return result


@numba.njit(cache=True, parallel=True)
def _fill_gauss_seidel_2d(
    a: np.ndarray,
    where: np.ndarray,
    num_iterations: int,
) -> np.ndarray:

    num_t, num_y, num_x = a.shape

    for t in numba.prange(num_t):
        for k in range(num_iterations):
            for is_odd in [False, True]:
                _iteration_gauss_seidel_2d(
                    a=a,
                    where=where,
                    t=t,
                    num_x=num_x,
                    num_y=num_y,
                    is_odd=is_odd,
                )

    return a


@numba.njit(cache=True, fastmath=True)
def _iteration_gauss_seidel_2d(
    a: np.ndarray,
    where: np.ndarray,
    t: int,
    num_x: int,
    num_y: int,
    is_odd: bool,
) -> None:

    xmin, xmax = -1, 1
    ymin, ymax = -1, 1

    dx = (xmax - xmin) / (num_x - 1)
    dy = (ymax - ymin) / (num_y - 1)

    dxxinv = 1 / (dx * dx)
    dyyinv = 1 / (dy * dy)

    dcent = 1 / (2 * (dxxinv + dyyinv))

    for j in range(num_y):
        for i in range(num_x):
            if (i + j) & 1 == is_odd:
                if where[t, j, i]:
                    i9 = (i - 1) % num_x
                    i1 = (i + 1) % num_x
                    j9 = (j - 1) % num_y
                    j1 = (j + 1) % num_y

                    xterm = dxxinv * (a[t, j, i9] + a[t, j, i1])
                    yterm = dyyinv * (a[t, j9, i] + a[t, j1, i])
                    a[t, j, i] = (xterm + yterm) * dcent



import pytest
import math
import regridding


@pytest.mark.parametrize(
    argnames="x,result_expected",
    argvalues=[
        (-2, -1),
        (-1, -1),
        (-0, 0),
        (0, 0),
        (1, 1),
        (0.5, 1),
        (math.nan, 0),
        (math.inf, 1),
        (-math.inf, -1),
    ],
)
def test_sign(
    x: float,
    result_expected: float,
):
    result = regridding.math.sign(x)
    assert result == result_expected



import pytest
import numpy as np
import scipy.ndimage
import regridding


sz_t = 9
sz_x = 10
sz_y = 11
sz_z = 12


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        np.random.random(sz_x),
    ],
)
@pytest.mark.parametrize(
    argnames="x",
    argvalues=[
        np.arange(sz_x),
        np.linspace(0, sz_x - 1, num=21),
    ],
)
@pytest.mark.parametrize(
    argnames="axis",
    argvalues=[None, 0, -1],
)
def test_ndarray_linear_interpolation_1d(
    a: np.ndarray,
    x: np.ndarray,
    axis: None | int | tuple[int],
):
    result = regridding.ndarray_linear_interpolation(a=a, indices=(x,), axis=axis)
    expected = scipy.ndimage.map_coordinates(
        input=a,
        coordinates=x[np.newaxis],
        mode="nearest",
        order=1,
    )

    assert np.allclose(result, expected)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        np.random.random((sz_x, sz_y)),
    ],
)
@pytest.mark.parametrize(
    argnames="x",
    argvalues=[
        np.arange(sz_x)[:, np.newaxis],
        np.linspace(0, sz_x - 1, num=100)[:, np.newaxis],
    ],
)
@pytest.mark.parametrize(
    argnames="y",
    argvalues=[
        np.arange(sz_y)[np.newaxis, :],
        np.linspace(0, sz_y - 1, num=5)[np.newaxis, :],
    ],
)
@pytest.mark.parametrize(
    argnames="axis",
    argvalues=[
        None,
        (0, 1),
        (0, ~0),
    ],
)
def test_ndarray_linear_interpolation_2d(
    a: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    axis: None | tuple[int],
):
    x, y = np.broadcast_arrays(x, y)

    result = regridding.ndarray_linear_interpolation(a=a, indices=(x, y), axis=axis)
    expected = scipy.ndimage.map_coordinates(
        input=a,
        coordinates=np.stack([x, y]),
        order=1,
    )

    assert np.allclose(result, expected)



import numpy as np


def _normalize_axis(
    axis: None | int | tuple[int, ...],
    ndim: int,
) -> tuple[int, ...]:
    if axis is None:
        axis = tuple(range(ndim))
    axis = np.core.numeric.normalize_axis_tuple(axis, ndim=ndim)
    axis = tuple(~(~np.array(axis) % ndim))
    return axis


def _normalize_input_output_coordinates(
    coordinates_input: tuple[np.ndarray, ...],
    coordinates_output: tuple[np.ndarray, ...],
    axis_input: None | int | tuple[int, ...] = None,
    axis_output: None | int | tuple[int, ...] = None,
) -> tuple[
    tuple[np.ndarray, ...],
    tuple[np.ndarray, ...],
    tuple[int, ...],
    tuple[int, ...],
    tuple[int, ...],
    tuple[int, ...],
    tuple[int, ...],
]:
    shape_coordinates_input = np.broadcast(*coordinates_input).shape
    shape_coordinates_output = np.broadcast(*coordinates_output).shape

    ndim_input = len(shape_coordinates_input)
    ndim_output = len(shape_coordinates_output)

    axis_input = _normalize_axis(axis_input, ndim=ndim_input)
    axis_output = _normalize_axis(axis_output, ndim=ndim_output)

    axis_input = sorted(axis_input, reverse=True)
    axis_output = sorted(axis_output, reverse=True)

    if len(axis_output) != len(axis_input):
        raise ValueError(
            f"The number of axes in `axis_output`, {axis_output}, "
            f"must match the number of axes in `axis_input`, {axis_input}"
        )

    if len(coordinates_input) != len(axis_input):
        raise ValueError(
            f"The number of elements in `coordinates_input`, {len(coordinates_input)}, "
            f"should match the number of axes in `axis_input`, {axis_input}"
        )

    if len(coordinates_output) != len(coordinates_input):
        raise ValueError(
            f"The number of elements in `coordinates_output`, {len(coordinates_output)}, "
            f"should match the number of elements in `coordinates_input`, {len(coordinates_input)}"
        )

    axis_input_orthogonal = tuple(
        ax for ax in _normalize_axis(None, ndim_input) if ax not in axis_input
    )
    axis_output_orthogonal = tuple(
        ax for ax in _normalize_axis(None, ndim_output) if ax not in axis_output
    )

    shape_input_orthogonal = tuple(
        shape_coordinates_input[ax] for ax in axis_input_orthogonal
    )
    shape_output_orthogonal = tuple(
        shape_coordinates_output[ax] for ax in axis_output_orthogonal
    )

    shape_orthogonal = np.broadcast_shapes(
        shape_input_orthogonal, shape_output_orthogonal
    )

    shape_input = list(shape_orthogonal)
    for ax in reversed(axis_input):
        ax = ax % ndim_input
        shape_input.insert(ax, shape_coordinates_input[ax])
    shape_input = tuple(shape_input)

    shape_output = list(shape_orthogonal)
    for ax in reversed(axis_output):
        ax = ax % ndim_input
        shape_output.insert(ax, shape_coordinates_output[ax])
    shape_output = tuple(shape_output)

    coordinates_input = tuple(
        np.broadcast_to(coord, shape_input) for coord in coordinates_input
    )
    coordinates_output = tuple(
        np.broadcast_to(coord, shape_output) for coord in coordinates_output
    )

    return (
        coordinates_input,
        coordinates_output,
        axis_input,
        axis_output,
        shape_input,
        shape_output,
        shape_orthogonal,
    )



from typing import Sequence, Literal
import numpy as np
from ._weights_multilinear import _weights_multilinear
from ._weights_conservative import _weights_conservative

__all__ = [
    "weights",
]


def weights(
    coordinates_input: tuple[np.ndarray, ...],
    coordinates_output: tuple[np.ndarray, ...],
    axis_input: None | int | Sequence[int] = None,
    axis_output: None | int | Sequence[int] = None,
    method: Literal["multilinear", "conservative"] = "multilinear",
) -> tuple[np.ndarray, tuple[int, ...], tuple[int, ...]]:
    """
    Save the results of a regridding operation as a sequence of weights,
    which can be used in subsequent regridding operations on the same grid.

    The results of this function are designed to be used by
    :func:`regridding.regrid_from_weights`

    This function returns a tuple containing a ragged array of weights,
    the shape of the input coordinates, and the shape of the output coordinates.

    Parameters
    ----------
    coordinates_input
        Coordinates of the input grid.
    coordinates_output
        Coordinates of the output grid.
        Should have the same number of coordinates as the input grid.
    axis_input
        Logical axes of the input grid to resample.
        If :obj:`None`, resample all the axes of the input grid.
        The number of axes should be equal to the number of
        coordinates in the input grid.
    axis_output
        Logical axes of the output grid corresponding to the resampled axes
        of the input grid.
        If :obj:`None`, all the axes of the output grid correspond to resampled
        axes in the input grid.
        The number of axes should be equal to the number of
        coordinates in the output grid.
    method
        The type of regridding to use.

    See Also
    --------
    :func:`regridding.regrid`
    :func:`regridding.regrid_from_weights`

    Examples
    --------

    Regrid two arrays of values defined on the same grid using saved weights.

    .. jupyter-execute::

        import numpy as np
        import scipy.signal
        import matplotlib.pyplot as plt
        import regridding

        # Define input grid
        x_input = np.linspace(-4, 4, num=101)
        y_input = np.linspace(-4, 4, num=101)
        x_input, y_input = np.meshgrid(x_input, y_input, indexing="ij")

        # Define rotated output grid
        angle = 0.2
        x_output = x_input * np.cos(angle) - y_input * np.sin(angle)
        y_output = x_input * np.sin(angle) + y_input * np.cos(angle)

        # Define two arrays of values defined on the same grid
        values_input_1 = np.cos(np.square(x_input)) * np.cos(np.square(y_input))
        values_input_2 = np.sin(np.square(x_input) + np.square(y_input))

        # Convolve with a 2x2 uniform kernel to simulate values defined on cell centers
        values_input_1 = scipy.signal.convolve(values_input_1, np.ones((2, 2)), mode="valid")
        values_input_2 = scipy.signal.convolve(values_input_2, np.ones((2, 2)), mode="valid")

        # Save regridding weights relating the input and output grids
        weights = regridding.weights(
            coordinates_input=(x_input, y_input),
            coordinates_output=(x_output, y_output),
            method="conservative",
        )

        # Regrid the first array of values using the saved weights
        values_output_1 = regridding.regrid_from_weights(
            *weights,
            values_input=values_input_1,
        )

        # Regrid the second array of values using the saved weights
        values_output_2 = regridding.regrid_from_weights(
            *weights,
            values_input=values_input_2,
        )

        # Plot the original and regridded arrays of values
        fig, axs = plt.subplots(
            nrows=2,
            ncols=2,
            sharex=True,
            sharey=True,
            constrained_layout=True,
        )
        axs[0, 0].pcolormesh(x_input, y_input, values_input_1);
        axs[0, 0].set_title(r"values_input_1");
        axs[0, 1].pcolormesh(x_input, y_input, values_input_2);
        axs[0, 1].set_title(r"values_input_2");
        axs[1, 0].pcolormesh(x_output, y_output, values_output_1);
        axs[1, 0].set_title(r"values_output_1");
        axs[1, 1].pcolormesh(x_output, y_output, values_output_2);
        axs[1, 1].set_title(r"values_output_2");
    """
    if method == "multilinear":
        return _weights_multilinear(
            coordinates_input=coordinates_input,
            coordinates_output=coordinates_output,
            axis_input=axis_input,
            axis_output=axis_output,
        )
    elif method == "conservative":
        return _weights_conservative(
            coordinates_input=coordinates_input,
            coordinates_output=coordinates_output,
            axis_input=axis_input,
            axis_output=axis_output,
        )
    else:
        raise ValueError(f"unrecognized method '{method}'")



from typing import Sequence, Literal
import numpy as np
from ._gauss_seidel import fill_gauss_seidel

__all__ = [
    "fill",
]


def fill(
    a: np.ndarray,
    where: None | np.ndarray = None,
    axis: None | int | Sequence[int] = None,
    method: Literal["gauss_seidel"] = "gauss_seidel",
    **kwargs,
) -> np.ndarray:
    """
    Fill an array with missing values by interpolating from the valid points.

    Parameters
    ----------
    a
        The array with missing values to be filled
    where
        Boolean array of missing values.
        If :obj:`None` (the default), all NaN values will be filled.
    axis
        The axes to use for interpolation.
        If :obj:`None` (the default), interpolate along all the axes of `a`.
    method
        The interpolation method to use.
        The only option is "gauss_seidel", which uses the Gauss-Seidel relaxation
        technique to interpolate the valid data points.
    kwargs
        Additional method-specific keyword arguments.
        For the Gauss-Seidel method, the valid keyword arguments are:
        - ``num_iterations=100``, the number of red-black Gauss-Seidel iterations to perform.

    Examples
    --------

    Set random elements of an array to NaN, and then fill in the missing elements
    using the Gauss-Seidel relaxation method.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import regridding

        # Define the independent variables
        x = 3 * np.pi * np.linspace(-1, 1, num=51)
        y = 3 * np.pi * np.linspace(-1, 1, num=51)
        x, y = np.meshgrid(x, y, indexing="ij")

        # Define the array to remove elements from
        a = np.cos(x) * np.cos(y)

        # Define the elements of the array to remove
        where = np.random.uniform(0, 1, size=a.shape) > 0.9

        # Set random elements of the array to NaN
        a_missing = a.copy()
        a_missing[where] = np.nan

        # Fill the missing elements using Gauss-Seidel relaxation
        b = regridding.fill(a_missing, method="gauss_seidel", num_iterations=11)

        # Plot the results
        fig, axs = plt.subplots(
            ncols=3,
            figsize=(6, 3),
            sharey=True,
            constrained_layout=True,
        )
        kwargs_imshow = dict(
            vmin=a.min(),
            vmax=a.max(),
        )
        axs[0].imshow(a_missing, **kwargs_imshow);
        axs[1].imshow(b, **kwargs_imshow);
        axs[2].imshow(a - b, **kwargs_imshow);
        axs[0].set_title("original array");
        axs[1].set_title("filled array");
        axs[2].set_title("difference");
    """

    if where is None:
        where = np.isnan(a)

    if method == "gauss_seidel":
        return fill_gauss_seidel(
            a=a,
            where=where,
            axis=axis,
            **kwargs,
        )
    else:  # pragma: nocover
        raise ValueError("Unrecognized method '{method}'")



import pytest
import numpy as np
import regridding

_num_x = 11
_num_y = 12
_num_t = 13


@pytest.mark.parametrize(
    argnames="a,where,axis",
    argvalues=[
        (
            np.random.uniform(0, 1, size=(_num_x, _num_y)),
            np.random.uniform(0, 1, size=(_num_x, _num_y)) > 0.9,
            None,
        ),
        (
            np.random.uniform(0, 1, size=(_num_t, _num_x, _num_y)),
            np.random.uniform(0, 1, size=(_num_t, _num_x, _num_y)) > 0.9,
            (~1, ~0),
        ),
        (
            np.sqrt(np.random.uniform(-0.1, 1, size=(_num_x, _num_t, _num_y))),
            None,
            (0, ~0),
        ),
    ],
)
@pytest.mark.parametrize("num_iterations", [11])
def test_fill_gauss_sidel_2d(
    a: np.ndarray,
    where: np.ndarray,
    axis: None | tuple[int, ...],
    num_iterations: int,
):
    result = regridding.fill(
        a=a,
        where=where,
        axis=axis,
        method="gauss_seidel",
        num_iterations=num_iterations,
    )
    if where is None:
        where = np.isnan(a)

    assert np.allclose(result[~where], a[~where])
    assert np.all(result[where] != 0)
